# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import milvus_pb2 as milvus__pb2
from . import status_pb2 as status__pb2


class MilvusServiceStub(object):
    # missing associated documentation comment in .proto file
    pass

    def __init__(self, channel):
        """Constructor.

        Args:
          channel: A grpc.Channel.
        """
        self.CreateCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/CreateCollection',
            request_serializer=milvus__pb2.CollectionSchema.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.HasCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/HasCollection',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.BoolReply.FromString,
        )
        self.DescribeCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/DescribeCollection',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.CollectionSchema.FromString,
        )
        self.CountCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/CountCollection',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.CollectionRowCount.FromString,
        )
        self.ShowCollections = channel.unary_unary(
            '/milvus.grpc.MilvusService/ShowCollections',
            request_serializer=milvus__pb2.Command.SerializeToString,
            response_deserializer=milvus__pb2.CollectionNameList.FromString,
        )
        self.ShowCollectionInfo = channel.unary_unary(
            '/milvus.grpc.MilvusService/ShowCollectionInfo',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.CollectionInfo.FromString,
        )
        self.DropCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/DropCollection',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.CreateIndex = channel.unary_unary(
            '/milvus.grpc.MilvusService/CreateIndex',
            request_serializer=milvus__pb2.IndexParam.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.DescribeIndex = channel.unary_unary(
            '/milvus.grpc.MilvusService/DescribeIndex',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.IndexParam.FromString,
        )
        self.DropIndex = channel.unary_unary(
            '/milvus.grpc.MilvusService/DropIndex',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.CreatePartition = channel.unary_unary(
            '/milvus.grpc.MilvusService/CreatePartition',
            request_serializer=milvus__pb2.PartitionParam.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.HasPartition = channel.unary_unary(
            '/milvus.grpc.MilvusService/HasPartition',
            request_serializer=milvus__pb2.PartitionParam.SerializeToString,
            response_deserializer=milvus__pb2.BoolReply.FromString,
        )
        self.ShowPartitions = channel.unary_unary(
            '/milvus.grpc.MilvusService/ShowPartitions',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.PartitionList.FromString,
        )
        self.DropPartition = channel.unary_unary(
            '/milvus.grpc.MilvusService/DropPartition',
            request_serializer=milvus__pb2.PartitionParam.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.Insert = channel.unary_unary(
            '/milvus.grpc.MilvusService/Insert',
            request_serializer=milvus__pb2.InsertParam.SerializeToString,
            response_deserializer=milvus__pb2.VectorIds.FromString,
        )
        self.GetVectorsByID = channel.unary_unary(
            '/milvus.grpc.MilvusService/GetVectorsByID',
            request_serializer=milvus__pb2.VectorsIdentity.SerializeToString,
            response_deserializer=milvus__pb2.VectorsData.FromString,
        )
        self.GetVectorIDs = channel.unary_unary(
            '/milvus.grpc.MilvusService/GetVectorIDs',
            request_serializer=milvus__pb2.GetVectorIDsParam.SerializeToString,
            response_deserializer=milvus__pb2.VectorIds.FromString,
        )
        self.Search = channel.unary_unary(
            '/milvus.grpc.MilvusService/Search',
            request_serializer=milvus__pb2.SearchParam.SerializeToString,
            response_deserializer=milvus__pb2.TopKQueryResult.FromString,
        )
        self.SearchByID = channel.unary_unary(
            '/milvus.grpc.MilvusService/SearchByID',
            request_serializer=milvus__pb2.SearchByIDParam.SerializeToString,
            response_deserializer=milvus__pb2.TopKQueryResult.FromString,
        )
        self.SearchInFiles = channel.unary_unary(
            '/milvus.grpc.MilvusService/SearchInFiles',
            request_serializer=milvus__pb2.SearchInFilesParam.SerializeToString,
            response_deserializer=milvus__pb2.TopKQueryResult.FromString,
        )
        self.Cmd = channel.unary_unary(
            '/milvus.grpc.MilvusService/Cmd',
            request_serializer=milvus__pb2.Command.SerializeToString,
            response_deserializer=milvus__pb2.StringReply.FromString,
        )
        self.DeleteByID = channel.unary_unary(
            '/milvus.grpc.MilvusService/DeleteByID',
            request_serializer=milvus__pb2.DeleteByIDParam.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.PreloadCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/PreloadCollection',
            request_serializer=milvus__pb2.PreloadCollectionParam.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.ReloadSegments = channel.unary_unary(
            '/milvus.grpc.MilvusService/ReloadSegments',
            request_serializer=milvus__pb2.ReLoadSegmentsParam.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.Flush = channel.unary_unary(
            '/milvus.grpc.MilvusService/Flush',
            request_serializer=milvus__pb2.FlushParam.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.Compact = channel.unary_unary(
            '/milvus.grpc.MilvusService/Compact',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.CreateHybridCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/CreateHybridCollection',
            request_serializer=milvus__pb2.Mapping.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.HasHybridCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/HasHybridCollection',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.BoolReply.FromString,
        )
        self.DropHybridCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/DropHybridCollection',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.DescribeHybridCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/DescribeHybridCollection',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.Mapping.FromString,
        )
        self.CountHybridCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/CountHybridCollection',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.CollectionRowCount.FromString,
        )
        self.ShowHybridCollections = channel.unary_unary(
            '/milvus.grpc.MilvusService/ShowHybridCollections',
            request_serializer=milvus__pb2.Command.SerializeToString,
            response_deserializer=milvus__pb2.MappingList.FromString,
        )
        self.ShowHybridCollectionInfo = channel.unary_unary(
            '/milvus.grpc.MilvusService/ShowHybridCollectionInfo',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=milvus__pb2.CollectionInfo.FromString,
        )
        self.PreloadHybridCollection = channel.unary_unary(
            '/milvus.grpc.MilvusService/PreloadHybridCollection',
            request_serializer=milvus__pb2.CollectionName.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )
        self.InsertEntity = channel.unary_unary(
            '/milvus.grpc.MilvusService/InsertEntity',
            request_serializer=milvus__pb2.HInsertParam.SerializeToString,
            response_deserializer=milvus__pb2.HEntityIDs.FromString,
        )
        self.HybridSearch = channel.unary_unary(
            '/milvus.grpc.MilvusService/HybridSearch',
            request_serializer=milvus__pb2.HSearchParam.SerializeToString,
            response_deserializer=milvus__pb2.TopKQueryResult.FromString,
        )
        self.HybridSearchInSegments = channel.unary_unary(
            '/milvus.grpc.MilvusService/HybridSearchInSegments',
            request_serializer=milvus__pb2.HSearchInSegmentsParam.SerializeToString,
            response_deserializer=milvus__pb2.TopKQueryResult.FromString,
        )
        self.GetEntityByID = channel.unary_unary(
            '/milvus.grpc.MilvusService/GetEntityByID',
            request_serializer=milvus__pb2.HEntityIdentity.SerializeToString,
            response_deserializer=milvus__pb2.HEntity.FromString,
        )
        self.GetEntityIDs = channel.unary_unary(
            '/milvus.grpc.MilvusService/GetEntityIDs',
            request_serializer=milvus__pb2.HGetEntityIDsParam.SerializeToString,
            response_deserializer=milvus__pb2.HEntityIDs.FromString,
        )
        self.DeleteEntitiesByID = channel.unary_unary(
            '/milvus.grpc.MilvusService/DeleteEntitiesByID',
            request_serializer=milvus__pb2.HDeleteByIDParam.SerializeToString,
            response_deserializer=status__pb2.Status.FromString,
        )


class MilvusServiceServicer(object):
    # missing associated documentation comment in .proto file
    pass

    def CreateCollection(self, request, context):
        """*
        @brief This method is used to create collection

        @param CollectionSchema, use to provide collection information to be created.

        @return Status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HasCollection(self, request, context):
        """*
        @brief This method is used to test collection existence.

        @param CollectionName, collection name is going to be tested.

        @return BoolReply
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DescribeCollection(self, request, context):
        """*
        @brief This method is used to get collection schema.

        @param CollectionName, target collection name.

        @return CollectionSchema
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CountCollection(self, request, context):
        """*
        @brief This method is used to get collection schema.

        @param CollectionName, target collection name.

        @return CollectionRowCount
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ShowCollections(self, request, context):
        """*
        @brief This method is used to list all collections.

        @param Command, dummy parameter.

        @return CollectionNameList
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ShowCollectionInfo(self, request, context):
        """*
        @brief This method is used to get collection detail information.

        @param CollectionName, target collection name.

        @return CollectionInfo
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropCollection(self, request, context):
        """*
        @brief This method is used to delete collection.

        @param CollectionName, collection name is going to be deleted.

        @return CollectionNameList
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateIndex(self, request, context):
        """*
        @brief This method is used to build index by collection in sync mode.

        @param IndexParam, index paramters.

        @return Status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DescribeIndex(self, request, context):
        """*
        @brief This method is used to describe index

        @param CollectionName, target collection name.

        @return IndexParam
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropIndex(self, request, context):
        """*
        @brief This method is used to drop index

        @param CollectionName, target collection name.

        @return Status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreatePartition(self, request, context):
        """*
        @brief This method is used to create partition

        @param PartitionParam, partition parameters.

        @return Status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HasPartition(self, request, context):
        """*
        @brief This method is used to test partition existence.

        @param PartitionParam, target partition.

        @return BoolReply
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ShowPartitions(self, request, context):
        """*
        @brief This method is used to show partition information

        @param CollectionName, target collection name.

        @return PartitionList
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropPartition(self, request, context):
        """*
        @brief This method is used to drop partition

        @param PartitionParam, target partition.

        @return Status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Insert(self, request, context):
        """*
        @brief This method is used to add vector array to collection.

        @param InsertParam, insert parameters.

        @return VectorIds
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetVectorsByID(self, request, context):
        """*
        @brief This method is used to get vectors data by id array.

        @param VectorsIdentity, target vector id array.

        @return VectorsData
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetVectorIDs(self, request, context):
        """*
        @brief This method is used to get vector ids from a segment

        @param GetVectorIDsParam, target collection and segment

        @return VectorIds
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Search(self, request, context):
        """*
        @brief This method is used to query vector in collection.

        @param SearchParam, search parameters.

        @return TopKQueryResult
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SearchByID(self, request, context):
        """*
        @brief This method is used to query vector by id.

        @param SearchByIDParam, search parameters.

        @return TopKQueryResult
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SearchInFiles(self, request, context):
        """*
        @brief This method is used to query vector in specified files.

        @param SearchInFilesParam, search in files paremeters.

        @return TopKQueryResult
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Cmd(self, request, context):
        """*
        @brief This method is used to give the server status.

        @param Command, command string

        @return StringReply
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteByID(self, request, context):
        """*
        @brief This method is used to delete vector by id

        @param DeleteByIDParam, delete parameters.

        @return status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PreloadCollection(self, request, context):
        """*
        @brief This method is used to preload collection/partitions

        @param PreloadCollectionParam, target collection/partitions.

        @return Status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReloadSegments(self, request, context):
        """*
        @brief This method is used to reload collection segments

        @param ReLoadSegmentsParam, target segments information.

        @return Status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Flush(self, request, context):
        """*
        @brief This method is used to flush buffer into storage.

        @param FlushParam, flush parameters

        @return Status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Compact(self, request, context):
        """*
        @brief This method is used to compact collection

        @param CollectionName, target collection name.

        @return Status
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateHybridCollection(self, request, context):
        """*******************************New Interface*******************************************

        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HasHybridCollection(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropHybridCollection(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DescribeHybridCollection(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CountHybridCollection(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ShowHybridCollections(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ShowHybridCollectionInfo(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PreloadHybridCollection(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InsertEntity(self, request, context):
        """/////////////////////////////////////////////////////////////////

        rpc CreateIndex(IndexParam) returns (Status) {}

        rpc DescribeIndex(CollectionName) returns (IndexParam) {}

        rpc DropIndex(CollectionName) returns (Status) {}

        /////////////////////////////////////////////////////////////////

        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HybridSearch(self, request, context):
        """TODO(yukun): will change to HQueryResult
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HybridSearchInSegments(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEntityByID(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEntityIDs(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteEntitiesByID(self, request, context):
        # missing associated documentation comment in .proto file
        pass
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MilvusServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'CreateCollection': grpc.unary_unary_rpc_method_handler(
            servicer.CreateCollection,
            request_deserializer=milvus__pb2.CollectionSchema.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'HasCollection': grpc.unary_unary_rpc_method_handler(
            servicer.HasCollection,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.BoolReply.SerializeToString,
        ),
        'DescribeCollection': grpc.unary_unary_rpc_method_handler(
            servicer.DescribeCollection,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.CollectionSchema.SerializeToString,
        ),
        'CountCollection': grpc.unary_unary_rpc_method_handler(
            servicer.CountCollection,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.CollectionRowCount.SerializeToString,
        ),
        'ShowCollections': grpc.unary_unary_rpc_method_handler(
            servicer.ShowCollections,
            request_deserializer=milvus__pb2.Command.FromString,
            response_serializer=milvus__pb2.CollectionNameList.SerializeToString,
        ),
        'ShowCollectionInfo': grpc.unary_unary_rpc_method_handler(
            servicer.ShowCollectionInfo,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.CollectionInfo.SerializeToString,
        ),
        'DropCollection': grpc.unary_unary_rpc_method_handler(
            servicer.DropCollection,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'CreateIndex': grpc.unary_unary_rpc_method_handler(
            servicer.CreateIndex,
            request_deserializer=milvus__pb2.IndexParam.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'DescribeIndex': grpc.unary_unary_rpc_method_handler(
            servicer.DescribeIndex,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.IndexParam.SerializeToString,
        ),
        'DropIndex': grpc.unary_unary_rpc_method_handler(
            servicer.DropIndex,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'CreatePartition': grpc.unary_unary_rpc_method_handler(
            servicer.CreatePartition,
            request_deserializer=milvus__pb2.PartitionParam.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'HasPartition': grpc.unary_unary_rpc_method_handler(
            servicer.HasPartition,
            request_deserializer=milvus__pb2.PartitionParam.FromString,
            response_serializer=milvus__pb2.BoolReply.SerializeToString,
        ),
        'ShowPartitions': grpc.unary_unary_rpc_method_handler(
            servicer.ShowPartitions,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.PartitionList.SerializeToString,
        ),
        'DropPartition': grpc.unary_unary_rpc_method_handler(
            servicer.DropPartition,
            request_deserializer=milvus__pb2.PartitionParam.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'Insert': grpc.unary_unary_rpc_method_handler(
            servicer.Insert,
            request_deserializer=milvus__pb2.InsertParam.FromString,
            response_serializer=milvus__pb2.VectorIds.SerializeToString,
        ),
        'GetVectorsByID': grpc.unary_unary_rpc_method_handler(
            servicer.GetVectorsByID,
            request_deserializer=milvus__pb2.VectorsIdentity.FromString,
            response_serializer=milvus__pb2.VectorsData.SerializeToString,
        ),
        'GetVectorIDs': grpc.unary_unary_rpc_method_handler(
            servicer.GetVectorIDs,
            request_deserializer=milvus__pb2.GetVectorIDsParam.FromString,
            response_serializer=milvus__pb2.VectorIds.SerializeToString,
        ),
        'Search': grpc.unary_unary_rpc_method_handler(
            servicer.Search,
            request_deserializer=milvus__pb2.SearchParam.FromString,
            response_serializer=milvus__pb2.TopKQueryResult.SerializeToString,
        ),
        'SearchByID': grpc.unary_unary_rpc_method_handler(
            servicer.SearchByID,
            request_deserializer=milvus__pb2.SearchByIDParam.FromString,
            response_serializer=milvus__pb2.TopKQueryResult.SerializeToString,
        ),
        'SearchInFiles': grpc.unary_unary_rpc_method_handler(
            servicer.SearchInFiles,
            request_deserializer=milvus__pb2.SearchInFilesParam.FromString,
            response_serializer=milvus__pb2.TopKQueryResult.SerializeToString,
        ),
        'Cmd': grpc.unary_unary_rpc_method_handler(
            servicer.Cmd,
            request_deserializer=milvus__pb2.Command.FromString,
            response_serializer=milvus__pb2.StringReply.SerializeToString,
        ),
        'DeleteByID': grpc.unary_unary_rpc_method_handler(
            servicer.DeleteByID,
            request_deserializer=milvus__pb2.DeleteByIDParam.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'PreloadCollection': grpc.unary_unary_rpc_method_handler(
            servicer.PreloadCollection,
            request_deserializer=milvus__pb2.PreloadCollectionParam.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'ReloadSegments': grpc.unary_unary_rpc_method_handler(
            servicer.ReloadSegments,
            request_deserializer=milvus__pb2.ReLoadSegmentsParam.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'Flush': grpc.unary_unary_rpc_method_handler(
            servicer.Flush,
            request_deserializer=milvus__pb2.FlushParam.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'Compact': grpc.unary_unary_rpc_method_handler(
            servicer.Compact,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'CreateHybridCollection': grpc.unary_unary_rpc_method_handler(
            servicer.CreateHybridCollection,
            request_deserializer=milvus__pb2.Mapping.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'HasHybridCollection': grpc.unary_unary_rpc_method_handler(
            servicer.HasHybridCollection,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.BoolReply.SerializeToString,
        ),
        'DropHybridCollection': grpc.unary_unary_rpc_method_handler(
            servicer.DropHybridCollection,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'DescribeHybridCollection': grpc.unary_unary_rpc_method_handler(
            servicer.DescribeHybridCollection,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.Mapping.SerializeToString,
        ),
        'CountHybridCollection': grpc.unary_unary_rpc_method_handler(
            servicer.CountHybridCollection,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.CollectionRowCount.SerializeToString,
        ),
        'ShowHybridCollections': grpc.unary_unary_rpc_method_handler(
            servicer.ShowHybridCollections,
            request_deserializer=milvus__pb2.Command.FromString,
            response_serializer=milvus__pb2.MappingList.SerializeToString,
        ),
        'ShowHybridCollectionInfo': grpc.unary_unary_rpc_method_handler(
            servicer.ShowHybridCollectionInfo,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=milvus__pb2.CollectionInfo.SerializeToString,
        ),
        'PreloadHybridCollection': grpc.unary_unary_rpc_method_handler(
            servicer.PreloadHybridCollection,
            request_deserializer=milvus__pb2.CollectionName.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
        'InsertEntity': grpc.unary_unary_rpc_method_handler(
            servicer.InsertEntity,
            request_deserializer=milvus__pb2.HInsertParam.FromString,
            response_serializer=milvus__pb2.HEntityIDs.SerializeToString,
        ),
        'HybridSearch': grpc.unary_unary_rpc_method_handler(
            servicer.HybridSearch,
            request_deserializer=milvus__pb2.HSearchParam.FromString,
            response_serializer=milvus__pb2.TopKQueryResult.SerializeToString,
        ),
        'HybridSearchInSegments': grpc.unary_unary_rpc_method_handler(
            servicer.HybridSearchInSegments,
            request_deserializer=milvus__pb2.HSearchInSegmentsParam.FromString,
            response_serializer=milvus__pb2.TopKQueryResult.SerializeToString,
        ),
        'GetEntityByID': grpc.unary_unary_rpc_method_handler(
            servicer.GetEntityByID,
            request_deserializer=milvus__pb2.HEntityIdentity.FromString,
            response_serializer=milvus__pb2.HEntity.SerializeToString,
        ),
        'GetEntityIDs': grpc.unary_unary_rpc_method_handler(
            servicer.GetEntityIDs,
            request_deserializer=milvus__pb2.HGetEntityIDsParam.FromString,
            response_serializer=milvus__pb2.HEntityIDs.SerializeToString,
        ),
        'DeleteEntitiesByID': grpc.unary_unary_rpc_method_handler(
            servicer.DeleteEntitiesByID,
            request_deserializer=milvus__pb2.HDeleteByIDParam.FromString,
            response_serializer=status__pb2.Status.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'milvus.grpc.MilvusService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))



import datetime
from urllib.parse import urlparse
import logging
import threading
import ujson

import grpc
from grpc._cython import cygrpc

from ..grpc_gen import milvus_pb2_grpc
from ..grpc_gen import milvus_pb2 as grpc_types
from .abstract import ConnectIntf, CollectionSchema, IndexParam, PartitionParam, TopKQueryResult, HEntitySet
from .prepare import Prepare
from .types import MetricType, Status
from .check import (
    int_or_str,
    is_legal_host,
    is_legal_port,
)

from .asynch import SearchFuture, InsertFuture, CreateIndexFuture, CompactFuture, FlushFuture

from .hooks import BaseSearchHook
from .client_hooks import SearchHook, HybridSearchHook
from .exceptions import ParamError, NotConnectError
from ..settings import DefaultConfig as config
from . import __version__

LOGGER = logging.getLogger(__name__)


def error_handler(*rargs):
    def wrapper(func):
        def handler(self, *args, **kwargs):
            record_dict = {}
            try:
                record_dict["API start"] = str(datetime.datetime.now())
                if self._pre_ping:
                    self.ping()
                record_dict["RPC start"] = str(datetime.datetime.now())
                return func(self, *args, **kwargs)
            except grpc.FutureTimeoutError as e:
                record_dict["RPC timeout"] = str(datetime.datetime.now())
                LOGGER.error("\nAddr [{}] {}\nRequest timeout: {}\n\t{}".format(self.server_address, func.__name__, e, record_dict))
                status = Status(Status.UNEXPECTED_ERROR, message='Request timeout')
                return status if not rargs else tuple([status]) + rargs
            except grpc.RpcError as e:
                record_dict["RPC error"] = str(datetime.datetime.now())
                LOGGER.error("\nAddr [{}] {}\nRpc error: {}\n\t{}".format(self.server_address, func.__name__, e, record_dict))
                status = Status(e.code(), message='Error occurred. {}'.format(e.details()))
                return status if not rargs else tuple([status]) + rargs
            except Exception as e:
                record_dict["Exception"] = str(datetime.datetime.now())
                LOGGER.error("\nAddr [{}] {}\nExcepted error: {}\n\t{}".format(self.server_address, func.__name__, e, record_dict))
                raise e

        return handler

    return wrapper


def set_uri(host, port, uri):
    if host is not None:
        _port = port if port is not None else config.GRPC_PORT
        _host = host
    elif port is None:
        try:
            _uri = urlparse(uri) if uri else urlparse(config.GRPC_URI)
            _host = _uri.hostname
            _port = _uri.port
        except (AttributeError, ValueError, TypeError) as e:
            raise ParamError("uri is illegal: {}".format(e))
    else:
        raise ParamError("Param is not complete. Please invoke as follow:\n"
                         "\t(host = ${HOST}, port = ${PORT})\n"
                         "\t(uri = ${URI})\n")

    if not is_legal_host(_host) or not is_legal_port(_port):
        raise ParamError("host or port is illeagl")

    return "{}:{}".format(str(_host), str(_port))


# def connect(addr, timeout):
#     channel = grpc.insecure_channel(
#         addr,
#         options=[(cygrpc.ChannelArgKey.max_send_message_length, -1),
#                  (cygrpc.ChannelArgKey.max_receive_message_length, -1)]
#     )
#     try:
#         ft = grpc.channel_ready_future(channel)
#         ft.result(timeout=timeout)
#         return True
#     except grpc.FutureTimeoutError:
#         raise NotConnectError('Fail connecting to server on {}. Timeout'.format(addr))
#     except grpc.RpcError as e:
#         raise NotConnectError("Connect error: <{}>".format(e))
#     # Unexpected error
#     except Exception as e:
#         raise NotConnectError("Error occurred when trying to connect server:\n"
#                               "\t<{}>".format(str(e)))
#     finally:
#         ft.cancel()
#         ft.__del__()
#         channel.__del__()


class GrpcHandler(ConnectIntf):
    def __init__(self, host=None, port=None, pre_ping=True, **kwargs):
        self._channel = None
        self._stub = None
        self._uri = None
        self.status = None
        self._connected = False
        self._pre_ping = pre_ping
        # if self._pre_ping:
        self._max_retry = kwargs.get("max_retry", 3)

        # record
        self._id = kwargs.get("conn_id", 0)

        # condition
        self._condition = threading.Condition()
        self._request_id = 0

        # client hook
        self._search_hook = SearchHook()
        self._hybrid_search_hook = HybridSearchHook()
        self._search_file_hook = SearchHook()

        # set server uri if object is initialized with parameter
        _uri = kwargs.get("uri", None)
        self._setup(host, port, _uri, pre_ping)

    def __str__(self):
        attr_list = ['%s=%r' % (key, value)
                     for key, value in self.__dict__.items() if not key.startswith('_')]
        return '<Milvus: {}>'.format(', '.join(attr_list))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def _setup(self, host, port, uri, pre_ping=False):
        """
        Create a grpc channel and a stub

        :raises: NotConnectError

        """
        self._uri = set_uri(host, port, uri)
        self._channel = grpc.insecure_channel(
            self._uri,
            options=[(cygrpc.ChannelArgKey.max_send_message_length, -1),
                     (cygrpc.ChannelArgKey.max_receive_message_length, -1),
                     ('grpc.enable_retries', 1),
                     ('grpc.keepalive_time_ms', 55000)]
                     # (b'grpc.enable_http_proxy', 0)]
        )
        self._stub = milvus_pb2_grpc.MilvusServiceStub(self._channel)
        self.status = Status()

    def _pre_request(self):
        if self._pre_ping:
            self.ping()

    def _get_request_id(self):
        with self._condition:
            _id = self._request_id
            self._request_id += 1
            return _id

    def set_hook(self, **kwargs):
        """
        specify client hooks.
        The client hooks are used in methods which interact with server.
        Use key-value method to set hooks. Supported hook setting currently is as follow.

            search hook,
            search-in-file hook

        """

        # config search hook
        _search_hook = kwargs.get('search', None)
        if _search_hook:
            if not isinstance(_search_hook, BaseSearchHook):
                raise ParamError("search hook must be a subclass of `BaseSearchHook`")

            self._search_hook = _search_hook

        _search_file_hook = kwargs.get('search_in_file', None)
        if _search_file_hook:
            if not isinstance(_search_file_hook, BaseSearchHook):
                raise ParamError("search hook must be a subclass of `BaseSearchHook`")

            self._search_file_hook = _search_file_hook

    def ping(self, timeout=30):
        ft = grpc.channel_ready_future(self._channel)
        retry = self._max_retry
        try:
            while retry > 0:
                try:
                    ft.result(timeout=timeout)
                    return True
                except:
                    retry -= 1
                    LOGGER.debug("Retry connect addr <{}> {} times".format(self._uri, self._max_retry - retry))
                    if retry > 0:
                        continue
                    else:
                        LOGGER.error("Retry to connect server {} failed.".format(self._uri))
                        raise
        except grpc.FutureTimeoutError:
            raise NotConnectError('Fail connecting to server on {}. Timeout'.format(self._uri))
        except grpc.RpcError as e:
            raise NotConnectError("Connect error: <{}>".format(e))
        # Unexpected error
        except Exception as e:
            raise NotConnectError("Error occurred when trying to connect server:\n"
                                  "\t<{}>".format(str(e)))

    @property
    def server_address(self):
        """
        Server network address
        """
        return self._uri

    def server_version(self, timeout=30):
        """
        Provide server version

        :return:
            Status: indicate if operation is successful

            str : Server version

        :rtype: (Status, str)
        """
        return self._cmd(cmd='version', timeout=timeout)

    def server_status(self, timeout=30):
        """
        Provide server status

        :return:
            Status: indicate if operation is successful

            str : Server version

        :rtype: (Status, str)
        """
        return self._cmd(cmd='status', timeout=timeout)

    @error_handler(None)
    def _cmd(self, cmd, timeout=30):
        cmd = Prepare.cmd(cmd)
        rf = self._stub.Cmd.future(cmd, wait_for_ready=True, timeout=timeout)
        response = rf.result()
        if response.status.error_code == 0:
            return Status(message='Success!'), response.string_reply

        return Status(code=response.status.error_code, message=response.status.reason), None

    @error_handler()
    def create_collection(self, collection_name, dimension, index_file_size, metric_type, param, timeout=30):
        """
        Create collection

        :type  param: dict or TableSchema
        :param param: Provide collection information to be created

                `example param={'collection_name': 'name',
                                'dimension': 16,
                                'index_file_size': 1024 (default)，
                                'metric_type': Metric_type.L2 (default)
                                }`

                `OR using Prepare.collection_schema to create param`

        :param timeout: timeout, The unit is seconds
        :type  timeout: double

        :return: Status, indicate if operation is successful
        :rtype: Status
        """

        collection_schema = Prepare.collection_schema(collection_name, dimension, index_file_size, metric_type, param)

        rf = self._stub.CreateCollection.future(collection_schema, wait_for_ready=True, timeout=timeout)
        status = rf.result()
        if status.error_code == 0:
            return Status(message='Create collection successfully!')

        LOGGER.error(status)
        return Status(code=status.error_code, message=status.reason)

    @error_handler()
    def create_hybrid_collection(self, collection_name, fields, timeout=30):
        collection_schema = Prepare.collection_hybrid_schema(collection_name, fields)
        response = self._stub.CreateHybridCollection(collection_schema)
        if response.error_code == 0:
            return Status(message='Create collection successfully!')

        return Status(code=response.error_code, message=response.reason)

    @error_handler(False)
    def has_collection(self, collection_name, timeout=30, **kwargs):
        """

        This method is used to test collection existence.

        :param collection_name: collection name is going to be tested.
        :type  collection_name: str
        :param timeout: time waiting for server response
        :type  timeout: int

        :return:
            Status: indicate if vectors inserted successfully
            bool if given collection_name exists

        """

        collection_name = Prepare.collection_name(collection_name)

        rf = self._stub.HasCollection.future(collection_name, wait_for_ready=True, timeout=timeout)
        reply = rf.result()
        if reply.status.error_code == 0:
            return Status(), reply.bool_reply

        return Status(code=reply.status.error_code, message=reply.status.reason), False

    @error_handler(None)
    def describe_collection(self, collection_name, timeout=30, **kwargs):
        """
        Show collection information

        :type  collection_name: str
        :param collection_name: which collection to be shown

        :returns: (Status, collection_schema)
            Status: indicate if query is successful
            collection_schema: return when operation is successful
        :rtype: (Status, TableSchema)
        """
        collection_name = Prepare.collection_name(collection_name)
        rf = self._stub.DescribeCollection.future(collection_name, wait_for_ready=True, timeout=timeout)
        response = rf.result()

        if response.status.error_code == 0:
            collection = CollectionSchema(
                collection_name=response.collection_name,
                dimension=response.dimension,
                index_file_size=response.index_file_size,
                metric_type=MetricType(response.metric_type)
            )
            return Status(message='Describe collection successfully!'), collection

        LOGGER.error(response.status)
        return Status(code=response.status.error_code, message=response.status.reason), None

    @error_handler(None)
    def count_collection(self, collection_name, timeout=30, **kwargs):
        """
        obtain vector number in collection

        :type  collection_name: str
        :param collection_name: target collection name.

        :returns:
            Status: indicate if operation is successful

            res: int, collection row count
        """

        collection_name = Prepare.collection_name(collection_name)

        rf = self._stub.CountCollection.future(collection_name, wait_for_ready=True, timeout=timeout)
        response = rf.result()
        if response.status.error_code == 0:
            return Status(message='Success!'), response.collection_row_count

        return Status(code=response.status.error_code, message=response.status.reason), None

    @error_handler([])
    def show_collections(self, timeout=30):
        """
        Show all collections information in database

        :return:
            Status: indicate if this operation is successful

            collections: list of collection names, return when operation
                    is successful
        :rtype:
            (Status, list[str])
        """

        cmd = Prepare.cmd('show_collections')
        rf = self._stub.ShowCollections.future(cmd, wait_for_ready=True, timeout=timeout)
        response = rf.result()
        if response.status.error_code == 0:
            return Status(message='Show collections successfully!'), \
                   [name for name in response.collection_names if len(name) > 0]
        return Status(response.status.error_code, message=response.status.reason), []

    @error_handler(None)
    def show_collection_info(self, collection_name, timeout=30):
        request = grpc_types.CollectionName(collection_name=collection_name)

        rf = self._stub.ShowCollectionInfo.future(request, wait_for_ready=True, timeout=timeout)
        response = rf.result()
        rpc_status = response.status

        if rpc_status.error_code == 0:
            json_info = response.json_info
            return Status(), {} if not json_info else ujson.loads(json_info)

        return Status(rpc_status.error_code, rpc_status.reason), None

    @error_handler()
    def preload_collection(self, collection_name, partition_tags=None, timeout=None):
        """
        Load collection to cache in advance

        :type collection_name: str
        :param collection_name: collection to preload

        :returns:
            Status:  indicate if invoke is successful
        """

        param = Prepare.preload_param(collection_name, partition_tags)
        status = self._stub.PreloadCollection.future(param, wait_for_ready=True, timeout=timeout).result()
        return Status(code=status.error_code, message=status.reason)

    @error_handler()
    def reload_segments(self, collection_name, segment_ids, timeout=30):
        file_ids = list(map(int_or_str, segment_ids))
        request = Prepare.reload_param(collection_name, file_ids)
        status = self._stub.ReloadSegments.future(request, wait_for_ready=True, timeout=timeout).result()
        return Status(code=status.error_code, message=status.reason)

    @error_handler()
    def drop_collection(self, collection_name, timeout=20):
        """
        Delete collection with collection_name

        :type  collection_name: str
        :param collection_name: Name of the collection being deleted

        :return: Status, indicate if operation is successful
        :rtype: Status
        """

        collection_name = Prepare.collection_name(collection_name)

        rf = self._stub.DropCollection.future(collection_name, wait_for_ready=True, timeout=timeout)
        status = rf.result()
        if status.error_code == 0:
            return Status(message='Delete collection successfully!')
        return Status(code=status.error_code, message=status.reason)

    @error_handler([])
    def insert(self, collection_name, records, ids=None, partition_tag=None, params=None, timeout=None, **kwargs):
        """
        Add vectors to collection

        :param ids: list of id
        :type  ids: list[int]

        :type  collection_name: str
        :param collection_name: collection name been inserted

        :type  records: list[list[float]]

                `example records: [[1.2345],[1.2345]]`

                `OR using Prepare.records`

        :param records: list of vectors been inserted

        :type partition_tag: str or None.
            If partition_tag is None, vectors will be inserted into collection rather than partitions.

        :param partition_tag: the tag string of collection

        :type

        :type  timeout: int
        :param timeout: time waiting for server response

        :returns:
            Status: indicate if vectors inserted successfully
            ids: list of id, after inserted every vector is given a id
        :rtype: (Status, list(int))
        """
        insert_param = kwargs.get('insert_param', None)

        if insert_param and not isinstance(insert_param, grpc_types.InsertParam):
            raise ParamError("The value of key 'insert_param' is invalid")

        body = insert_param if insert_param \
            else Prepare.insert_param(collection_name, records, partition_tag, ids, params)

        rf = self._stub.Insert.future(body, wait_for_ready=True, timeout=timeout)
        if kwargs.get("_async", False) is True:
            cb = kwargs.get("_callback", None)
            return InsertFuture(rf, cb)

        response = rf.result()
        if response.status.error_code == 0:
            return Status(message='Add vectors successfully!'), list(response.vector_id_array)

        return Status(code=response.status.error_code, message=response.status.reason), []

    @error_handler([])
    def insert_hybrid(self, collection_name, entities, vector_entities, ids=None, tag=None, params=None):
        insert_param = Prepare.insert_hybrid_param(collection_name, tag, entities, vector_entities, ids, params)
        response = self._stub.InsertEntity(insert_param)
        status = response.status
        if status.error_code == 0:
            return Status(message='Insert successfully!'), list(response.entity_id_array)
        return Status(code=status.error_code, message=status.reason), []

    @error_handler([])
    def get_vectors_by_ids(self, collection_name, ids, timeout=30):
        request = grpc_types.VectorsIdentity(collection_name=collection_name, id_array=ids)

        rf = self._stub.GetVectorsByID.future(request, wait_for_ready=True, timeout=timeout)
        response = rf.result()
        status = response.status
        if status.error_code == 0:
            status = Status(message="Obtain vector successfully")
            vectors = list()
            for datas in response.vectors_data:
                vector = bytes(datas.binary_data) or list(datas.float_data)
                vectors.append(vector)
            return status, vectors

        return Status(code=status.error_code, message=status.reason), []

    @error_handler([])
    def get_hybrid_entity_by_id(self, collection_name, ids):
        request = grpc_types.VectorsIdentity(collection_name=collection_name, id_array=ids)
        response = self._stub.GetEntityByID(request)

        if response.status.error_code == 0:
            return Status(), HEntitySet(response)
        return Status(response.status.error_code, response.status.reason), []

    @error_handler([])
    def get_vector_ids(self, collection_name, segment_name, timeout=30):
        request = grpc_types.GetVectorIDsParam(collection_name=collection_name, segment_name=segment_name)

        rf = self._stub.GetVectorIDs.future(request, wait_for_ready=True, timeout=timeout)
        response = rf.result()

        if response.status.error_code == 0:
            return Status(), list(response.vector_id_array)
        return Status(response.status.error_code, response.status.reason), []

    @error_handler()
    def create_index(self, collection_name, index_type=None, params=None, timeout=None, **kwargs):
        """
        build vectors of specific collection and create vector index

        :param collection_name: collection used to crete index.
        :type collection_name: str
        :param index: index params
        :type index: dict

            index_param can be None

            `example (default) param={'index_type': IndexType.FLAT,
                            'nlist': 16384}`

        :param timeout: grpc request timeout.

            if `timeout` = -1, method invoke a synchronous call, waiting util grpc response
            else method invoke a asynchronous call, timeout work here

        :type  timeout: int

        :return: Status, indicate if operation is successful
        """
        index_param = Prepare.index_param(collection_name, index_type, params)
        # status = self._stub.CreateIndex.future(index_param).result(timeout=timeout)
        future = self._stub.CreateIndex.future(index_param, wait_for_ready=True, timeout=timeout)
        if kwargs.get('_async', False):
            cb = kwargs.get("_callback", None)
            return CreateIndexFuture(future, cb)
        status = future.result()

        if status.error_code == 0:
            return Status(message='Build index successfully!')
        return Status(code=status.error_code, message=status.reason)

    @error_handler()
    def describe_index(self, collection_name, timeout=30):
        """
        Show index information of designated collection

        :type collection_name: str
        :param collection_name: collection name been queried

        :returns:
            Status:  indicate if query is successful
            IndexSchema:

        """

        collection_name = Prepare.collection_name(collection_name)

        rf = self._stub.DescribeIndex.future(collection_name, wait_for_ready=True, timeout=timeout)
        index_param = rf.result()

        status = index_param.status

        if status.error_code == 0:
            return Status(message="Successfully"), \
                   IndexParam(index_param.collection_name,
                              index_param.index_type,
                              index_param.extra_params[0].value)

        return Status(code=status.error_code, message=status.reason), None

    @error_handler()
    def drop_index(self, collection_name, timeout=30):
        """
        drop index from index file

        :param collection_name: target collection name.
        :type collection_name: str

        :return:
            Status: indicate if operation is successful

        ：:rtype: Status
        """

        collection_name = Prepare.collection_name(collection_name)
        rf = self._stub.DropIndex.future(collection_name, wait_for_ready=True, timeout=timeout)
        status = rf.result()
        # status = self._stub.DropIndex.future(collection_name).result(timeout=timeout)
        return Status(code=status.error_code, message=status.reason)

    @error_handler()
    def create_partition(self, collection_name, partition_tag, timeout=30):
        """
        create a specific partition under designated collection. After done, the meta file in
        milvus server update partition information, you can perform actions about partitions
        with partition tag.

        :param collection_name: target collection name.
        :type  collection_name: str

        :param partition_name: name of target partition under designated collection.
        :type  partition_name: str

        :param partition_tag: tag name of target partition under designated collection.
        :type  partition_tag: str

        :param timeout: time waiting for response.
        :type  timeout: int

        :return:
            Status: indicate if operation is successful

        """
        request = Prepare.partition_param(collection_name, partition_tag)
        rf = self._stub.CreatePartition.future(request, wait_for_ready=True, timeout=timeout)
        response = rf.result()
        return Status(code=response.error_code, message=response.reason)

    @error_handler(False)
    def has_partition(self, collection_name, partition_tag, timeout=30):
        request = Prepare.partition_param(collection_name, partition_tag)
        rf = self._stub.HasPartition.future(request, wait_for_ready=True, timeout=timeout)
        response = rf.result()
        status = response.status
        return Status(code=status.error_code, message=status.reason), response.bool_reply

    @error_handler([])
    def show_partitions(self, collection_name, timeout=30):
        """
        Show all partitions under designated collection.

        :param collection_name: target collection name.
        :type  collection_name: str

        :param timeout: time waiting for response.
        :type  timeout: int

        :return:
            Status: indicate if operation is successful
            partition_list:

        """
        request = Prepare.collection_name(collection_name)

        rf = self._stub.ShowPartitions.future(request, wait_for_ready=True, timeout=timeout)
        response = rf.result()
        # response = self._stub.ShowPartitions.future(request).result(timeout=timeout)
        status = response.status
        if status.error_code == 0:
            partition_list = [PartitionParam(collection_name, p) for p in response.partition_tag_array]
            return Status(), partition_list

        return Status(code=status.error_code, message=status.reason), []

    @error_handler()
    def drop_partition(self, collection_name, partition_tag, timeout=30):
        """
        Drop specific partition under designated collection.

        :param collection_name: target collection name.
        :type  collection_name: str

        :param partition_tag: tag name of specific partition
        :type  partition_tag: str

        :param timeout: time waiting for response.
        :type  timeout: int

        :return:
            Status: indicate if operation is successful

        """
        request = grpc_types.PartitionParam(collection_name=collection_name, tag=partition_tag)

        rf = self._stub.DropPartition.future(request, wait_for_ready=True, timeout=timeout)
        response = rf.result()
        # response = self._stub.DropPartition.future(request).result(timeout=timeout)
        return Status(code=response.error_code, message=response.reason)

    @error_handler(None)
    def search(self, collection_name, top_k, query_records, partition_tags=None, params=None, timeout=None, **kwargs):
        request = Prepare.search_param(collection_name, top_k, query_records, partition_tags, params)

        self._search_hook.pre_search()
        if kwargs.get("_async", False) is True:
            future = self._stub.Search.future(request, wait_for_ready=True, timeout=timeout)

            func = kwargs.get("_callback", None)
            return SearchFuture(future, func)
        ft = self._stub.Search.future(request)
        response = ft.result()
        self._search_hook.aft_search()

        if self._search_hook.on_response():
            return response

        if response.status.error_code != 0:
            return Status(code=response.status.error_code,
                          message=response.status.reason), []

        resutls = self._search_hook.handle_response(response)
        return Status(message='Search vectors successfully!'), resutls

    @error_handler(None)
    def search_hybrid_pb(self, collection_name, query_entities, partition_tags, params=None, **kwargs):
        request = Prepare.search_hybrid_pb_param(collection_name, query_entities, partition_tags, params)
        response = self._stub.HybridSearchPB(request)

        if response.status.error_code != 0:
            return Status(code=response.status.error_code,
                          message=response.status.reason), []

        return Status(message='Search vectors successfully!'), \
               self._hybrid_search_hook.handle_response(response)

    @error_handler(None)
    def search_hybrid(self, collection_name, vector_params, dsl, partition_tags=None, params=None, **kwargs):
        request = Prepare.search_hybrid_param(collection_name, vector_params, dsl, partition_tags, params)
        response = self._stub.HybridSearch(request)

        if response.status.error_code != 0:
            return Status(code=response.status.error_code,
                          message=response.status.reason), []

        return Status(message='Search vectors successfully!'), \
               self._hybrid_search_hook.handle_response(response)

    @error_handler(None)
    def search_by_ids(self, collection_name, ids, top_k, partition_tags=None, params=None, timeout=None, **kwargs):
        request = Prepare.search_by_ids_param(collection_name, ids, top_k, partition_tags, params)
        if kwargs.get("_async", False) is True:
            future = self._stub.SearchByID.future(request, wait_for_ready=True, timeout=timeout)

            func = kwargs.get("_callback", None)
            return SearchFuture(future, func)

        ft = self._stub.SearchByID.future(request, wait_for_ready=True, timeout=timeout)
        response = ft.result()
        self._search_hook.aft_search()

        if self._search_hook.on_response():
            return response

        if response.status.error_code != 0:
            return Status(code=response.status.error_code,
                          message=response.status.reason), []

        return Status(message='Search vectors successfully!'), \
               self._search_hook.handle_response(response)

    @error_handler(None)
    def search_in_files(self, collection_name, file_ids, query_records, top_k, params, timeout=None, **kwargs):
        """
        Query vectors in a collection, in specified files.

        The server store vector data into multiple files if the size of vectors
        exceeds file size threshold. It is supported to search in several files
        by specifying file ids. However, these file ids are stored in db in server,
        and python sdk doesn't apply any APIs get them at client. It's a specific
        method used in shards. Obtain more detail about milvus shards, see
        <a href="https://github.com/milvus-io/milvus/tree/0.6.0/shards">

        :type  collection_name: str
        :param collection_name: collection name been queried

        :type  file_ids: list[str] or list[int]
        :param file_ids: Specified files id array

        :type  query_records: list[list[float]]
        :param query_records: all vectors going to be queried

        :param query_ranges: Optional ranges for conditional search.
            If not specified, search in the whole collection

        :type  top_k: int
        :param top_k: how many similar vectors will be searched

        :returns:
            Status:  indicate if query is successful
            results: query result

        :rtype: (Status, TopKQueryResult)
        """

        file_ids = list(map(int_or_str, file_ids))
        infos = Prepare.search_vector_in_files_param(
            collection_name, query_records, top_k, file_ids, params
        )

        self._search_file_hook.pre_search()

        if kwargs.get("_async", False) is True:
            future = self._stub.SearchInFiles.future(infos, wait_for_ready=True, timeout=timeout)

            func = kwargs.get("_callback", None)
            return SearchFuture(future, func)

        ft = self._stub.SearchInFiles.future(infos, wait_for_ready=True, timeout=timeout)
        response = ft.result()
        self._search_file_hook.aft_search()

        if self._search_file_hook.on_response():
            return response

        if response.status.error_code != 0:
            return Status(code=response.status.error_code,
                          message=response.status.reason), []

        return Status(message='Search vectors successfully!'), \
               self._search_file_hook.handle_response(response)

    @error_handler()
    def delete_by_id(self, collection_name, id_array, timeout=None):
        request = Prepare.delete_by_id_param(collection_name, id_array)

        rf = self._stub.DeleteByID.future(request, wait_for_ready=True, timeout=timeout)
        status = rf.result()
        # status = self._stub.DeleteByID.future(request).result(timeout=timeout)
        return Status(code=status.error_code, message=status.reason)

    @error_handler()
    def flush(self, collection_name_array, timeout=None, **kwargs):
        request = Prepare.flush_param(collection_name_array)
        future = self._stub.Flush.future(request, wait_for_ready=True, timeout=timeout)
        if kwargs.get("_async", False):
            cb = kwargs.get("_callback", None)
            return FlushFuture(future, cb)
        response = future.result()
        return Status(code=response.error_code, message=response.reason)

    @error_handler()
    def compact(self, collection_name, timeout, **kwargs):
        request = Prepare.compact_param(collection_name)
        future = self._stub.Compact.future(request, wait_for_ready=True, timeout=timeout)
        if kwargs.get("_async", False):
            cb = kwargs.get("_callback", None)
            return CompactFuture(future, cb)
        response = future.result()
        return Status(code=response.error_code, message=response.reason)

    def set_config(self, parent_key, child_key, value):
        cmd = "set_config {}.{} {}".format(parent_key, child_key, value)
        return self._cmd(cmd)

    def get_config(self, parent_key, child_key):
        cmd = "get_config {}.{}".format(parent_key, child_key)
        return self._cmd(cmd)



from .hooks import BaseSearchHook
from .abstract import TopKQueryResult, HybridResult


class SearchHook(BaseSearchHook):

    def handle_response(self, _response):
        """
        use class `TopKQueryResult` to deal with response from server
        """
        return TopKQueryResult(_response)


class HybridSearchHook(BaseSearchHook):

    def handle_response(self, _response):
        return HybridResult(_response)



# -*- coding: UTF-8 -*-

import collections
import copy
import functools
import logging
import threading

from urllib.parse import urlparse

from . import __version__
from .types import IndexType, MetricType, Status
from .check import check_pass_param, is_legal_host, is_legal_port
from .pool import ConnectionPool, SingleConnectionPool, SingletonThreadPool
from .grpc_handler import GrpcHandler
from .http_handler import HttpHandler
from .exceptions import ParamError, NotConnectError, DeprecatedError

from ..settings import DefaultConfig as config

LOGGER = logging.getLogger(__name__)


def deprecated(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        error_str = "Function {} has been deprecated".format(func.__name__)
        LOGGER.error(error_str)
        raise DeprecatedError(error_str)

    return inner


def check_connect(func):
    @functools.wraps(func)
    def inner(self, *args, **kwargs):
        return func(self, *args, **kwargs)

    return inner


def _pool_args(**kwargs):
    pool_kwargs = dict()
    for k, v in kwargs.items():
        if k in ("pool_size", "wait_timeout", "handler", "try_connect", "pre_ping", "max_retry"):
            pool_kwargs[k] = v

    return pool_kwargs


def _set_uri(host, port, uri, handler="GRPC"):
    default_port = config.GRPC_PORT if handler == "GRPC" else config.HTTP_PORT
    default_uri = config.GRPC_URI if handler == "GRPC" else config.HTTP_URI
    uri_prefix = "tcp://" if handler == "GRPC" else "http://"

    if host is not None:
        _port = port if port is not None else default_port
        _host = host
    elif port is None:
        try:
            _uri = urlparse(uri) if uri else urlparse(default_uri)
            _host = _uri.hostname
            _port = _uri.port
        except (AttributeError, ValueError, TypeError) as e:
            raise ParamError("uri is illegal: {}".format(e))
    else:
        raise ParamError("Param is not complete. Please invoke as follow:\n"
                         "\t(host = ${HOST}, port = ${PORT})\n"
                         "\t(uri = ${URI})\n")

    if not is_legal_host(_host) or not is_legal_port(_port):
        raise ParamError("host {} or port {} is illegal".format(_host, _port))

    return "{}{}:{}".format(uri_prefix, str(_host), str(_port))


class Milvus:
    def __init__(self, host=None, port=None, handler="GRPC", pool="SingletonThread", **kwargs):
        self._name = kwargs.get('name', None)
        self._uri = None
        self._status = None
        self._connected = False
        self._handler = handler

        _uri = kwargs.get('uri', None)
        pool_uri = _set_uri(host, port, _uri, self._handler)
        pool_kwargs = _pool_args(handler=handler, **kwargs)
        # self._pool = SingleConnectionPool(pool_uri, **pool_kwargs)
        if pool == "QueuePool":
            self._pool = ConnectionPool(pool_uri, **pool_kwargs)
        elif pool == "SingletonThread":
            self._pool = SingletonThreadPool(pool_uri, **pool_kwargs)
        elif pool == "Singleton":
            self._pool = SingleConnectionPool(pool_uri, **pool_kwargs)
        else:
            raise ParamError("Unknown pool value: {}".format(pool))

        # store extra key-words arguments
        self._kw = kwargs
        self._hooks = collections.defaultdict()

    def __enter__(self):
        self._conn = self._pool.fetch()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._conn.close()
        self._conn = None

    def __del__(self):
        return self.close()

    def _connection(self):
        return self._pool.fetch()

    @deprecated
    def set_hook(self, **kwargs):
        """
        Deprecated
        """
        # TODO: may remove it.
        if self._stub:
            return self._stub.set_hook(**kwargs)

        self._hooks.update(kwargs)

    @property
    def name(self):
        return self._name

    @property
    def handler(self):
        return self._handler

    @deprecated
    def connect(self, host=None, port=None, uri=None, timeout=2):
        """
        Deprecated
        """
        if self.connected() and self._connected:
            return Status(message="You have already connected {} !".format(self._uri),
                          code=Status.CONNECT_FAILED)

        if self._stub is None:
            self._init(host, port, uri, handler=self._handler)

        if self.ping(timeout):
            self._status = Status(message="Connected")
            self._connected = True
            return self._status

    @deprecated
    def connected(self):
        """
        Deprecated
        """
        return True if self._status and self._status.OK() else False

    @deprecated
    def disconnect(self):
        """
        Deprecated
        """
        pass

    def close(self):
        """
        Close client instance
        """
        self._pool = None

    def client_version(self):
        """
        Returns the version of the client.

        :return: Version of the client.

        :rtype: (str)
        """
        return __version__

    def server_status(self, timeout=30):
        """
        Returns the status of the Milvus server.

        :return:
            Status: Whether the operation is successful.

            str : Status of the Milvus server.

        :rtype: (Status, str)
        """
        return self._cmd("status", timeout)

    def server_version(self, timeout=30):
        """
        Returns the version of the Milvus server.

        :return:
           Status: Whether the operation is successful.

           str : Version of the Milvus server.

        :rtype: (Status, str)
        """

        return self._cmd("version", timeout)

    @check_connect
    def _cmd(self, cmd, timeout=30):
        check_pass_param(cmd=cmd)

        with self._connection() as handler:
            return handler._cmd(cmd, timeout)

    @check_connect
    def create_collection(self, param, timeout=30):
        """
        Creates a collection.

        :type  param: dict
        :param param: Information needed to create a collection.

                `param={'collection_name': 'name',
                                'dimension': 16,
                                'index_file_size': 1024 (default)，
                                'metric_type': Metric_type.L2 (default)
                                }`

        :param timeout: Timeout in seconds.
        :type  timeout: double

        :return: Whether the operation is successful.
        :rtype: Status
        """
        if not isinstance(param, dict):
            raise ParamError('Param type incorrect, expect {} but get {} instead'
                             .format(type(dict), type(param)))

        collection_param = copy.deepcopy(param)

        if 'collection_name' not in collection_param:
            raise ParamError('collection_name is required')
        collection_name = collection_param["collection_name"]
        collection_param.pop('collection_name')

        if 'dimension' not in collection_param:
            raise ParamError('dimension is required')
        dim = collection_param["dimension"]
        collection_param.pop("dimension")

        index_file_size = collection_param.get('index_file_size', 1024)
        collection_param.pop('index_file_size', None)

        metric_type = collection_param.get('metric_type', MetricType.L2)
        collection_param.pop('metric_type', None)

        check_pass_param(collection_name=collection_name, dimension=dim, index_file_size=index_file_size,
                         metric_type=metric_type)

        with self._connection() as handler:
            return handler.create_collection(collection_name, dim, index_file_size, metric_type, collection_param, timeout)

    @check_connect
    def create_hybrid_collection(self, collection_name, fields, timeout=30):
        with self._connection() as handler:
            return handler.create_hybrid_collection(collection_name, fields, timeout)

    @check_connect
    def has_collection(self, collection_name, timeout=30):
        """

        Checks whether a collection exists.

        :param collection_name: Name of the collection to check.
        :type  collection_name: str
        :param timeout: Timeout in seconds.
        :type  timeout: int

        :return:
            Status: indicate whether the operation is successful.
            bool if given collection_name exists

        """
        check_pass_param(collection_name=collection_name)
        with self._connection() as handler:
            return handler.has_collection(collection_name, timeout)

    @check_connect
    def get_collection_info(self, collection_name, timeout=30):
        """
        Returns information of a collection.

        :type  collection_name: str
        :param collection_name: Name of the collection to describe.

        :returns: (Status, table_schema)
            Status: indicate if query is successful
            table_schema: return when operation is successful

        :rtype: (Status, TableSchema)
        """
        check_pass_param(collection_name=collection_name)
        with self._connection() as handler:
            return handler.describe_collection(collection_name, timeout)

    @check_connect
    def count_entities(self, collection_name, timeout=30):
        """
        Returns the number of vectors in a collection.

        :type  collection_name: str
        :param collection_name: target table name.

        :returns:
            Status: indicate if operation is successful

            res: int, table row count
        """
        check_pass_param(collection_name=collection_name)
        with self._connection() as handler:
            return handler.count_collection(collection_name, timeout)

    @check_connect
    def list_collections(self, timeout=30):
        """
        Returns collection list.

        :return:
            Status: indicate if this operation is successful

            collections: list of collection names, return when operation
                    is successful
        :rtype:
            (Status, list[str])
        """
        with self._connection() as handler:
            return handler.show_collections(timeout)

    @check_connect
    def get_collection_stats(self, collection_name, timeout=30):
        """
        Returns collection statistics information

        :return:
            Status: indicate if this operation is successful

            statistics: statistics information
        :rtype:
            (Status, dict)
        """
        check_pass_param(collection_name=collection_name)
        with self._connection() as handler:
            return handler.show_collection_info(collection_name, timeout)

    @check_connect
    def load_collection(self, collection_name, partition_tags=None, timeout=None):
        """
        Loads a collection for caching.

        :type collection_name: str
        :param collection_name: collection to load

        :returns:
            Status:  indicate if invoke is successful
        """
        check_pass_param(collection_name=collection_name)
        with self._connection() as handler:
            return handler.preload_collection(collection_name, partition_tags, timeout)

    @check_connect
    def reload_segments(self, collection_name, segment_ids):
        check_pass_param(collection_name=collection_name)
        with self._connection() as handler:
            return handler.reload_segments(collection_name, segment_ids)

    @check_connect
    def drop_collection(self, collection_name, timeout=30):
        """
        Deletes a collection by name.

        :type  collection_name: str
        :param collection_name: Name of the collection being deleted

        :return: Status, indicate if operation is successful
        :rtype: Status
        """
        check_pass_param(collection_name=collection_name)
        with self._connection() as handler:
            return handler.drop_collection(collection_name, timeout)

    @check_connect
    def insert(self, collection_name, records, ids=None, partition_tag=None, params=None, timeout=None, **kwargs):
        """
        Insert vectors to a collection.

        :param ids: list of id
        :type  ids: list[int]

        :type  collection_name: str
        :param collection_name: Name of the collection to insert vectors to.

        :type  records: list[list[float]]

                `example records: [[1.2345],[1.2345]]`

                `OR using Prepare.records`

        :param records: List of vectors to insert.

        :type partition_tag: str or None.
            If partition_tag is None, vectors will be inserted to the collection rather than partitions.

        :param partition_tag: Tag of a partition.

        :returns:
            Status: Whether vectors are inserted successfully.
            ids: IDs of the inserted vectors.
        :rtype: (Status, list(int))
        """
        if kwargs.get("insert_param", None) is not None:
            with self._connection() as handler:
                return handler.insert(None, None, timeout=timeout, **kwargs)

        check_pass_param(collection_name=collection_name, records=records)
        partition_tag is not None and check_pass_param(partition_tag=partition_tag)
        if ids is not None:
            check_pass_param(ids=ids)
            if len(records) != len(ids):
                raise ParamError("length of vectors do not match that of ids")

        params = params or dict()
        if not isinstance(params, dict):
            raise ParamError("Params must be a dictionary type")
        with self._connection() as handler:
            return handler.insert(collection_name, records, ids, partition_tag, params, timeout, **kwargs)

    @check_connect
    def insert_hybrid(self, collection_name, entities, vector_entities, ids=None, partition_tag=None, params=None):
        with self._connection() as handler:
            return handler.insert_hybrid(collection_name, entities, vector_entities, ids, partition_tag, params)

    def get_entity_by_id(self, collection_name, ids, timeout=None):
        """
        Returns raw vectors according to ids.

        :param collection_name: Name of the collection
        :type collection_name: str

        :param ids: list of vector id
        :type ids: list

        :returns:
            Status: indicate if invoke is successful

        """
        check_pass_param(collection_name=collection_name, ids=ids)

        with self._connection() as handler:
            return handler.get_vectors_by_ids(collection_name, ids, timeout=timeout)

    def get_hybrid_entity_by_id(self, collection_name, ids):
        check_pass_param(collection_name=collection_name, ids=ids)
        with self._connection() as handler:
            return handler.get_hybrid_entity_by_id(collection_name, ids)

    @check_connect
    def list_id_in_segment(self, collection_name, segment_name, timeout=None):
        check_pass_param(collection_name=collection_name)
        check_pass_param(collection_name=segment_name)
        with self._connection() as handler:
            return handler.get_vector_ids(collection_name, segment_name, timeout)

    @check_connect
    def create_index(self, collection_name, index_type=None, params=None, timeout=None, **kwargs):
        """
        Creates index for a collection.

        :param collection_name: Collection used to create index.
        :type collection_name: str
        :param index: index params
        :type index: dict

            index_param can be None

            `example (default) param={'index_type': IndexType.FLAT,
                            'nlist': 16384}`

        :param timeout: grpc request timeout.

            if `timeout` = -1, method invoke a synchronous call, waiting util grpc response
            else method invoke a asynchronous call, timeout work here

        :type  timeout: int

        :return: Whether the operation is successful.
        """
        _index_type = IndexType.FLAT if index_type is None else index_type
        check_pass_param(collection_name=collection_name, index_type=_index_type)

        params = params or dict()
        if not isinstance(params, dict):
            raise ParamError("Params must be a dictionary type")
        with self._connection() as handler:
            return handler.create_index(collection_name, _index_type, params, timeout, **kwargs)

    @check_connect
    def get_index_info(self, collection_name, timeout=30):
        """
        Show index information of a collection.

        :type collection_name: str
        :param collection_name: table name been queried

        :returns:
            Status:  Whether the operation is successful.
            IndexSchema:

        """
        check_pass_param(collection_name=collection_name)

        with self._connection() as handler:
            return handler.describe_index(collection_name, timeout)

    @check_connect
    def drop_index(self, collection_name, timeout=30):
        """
        Removes an index.

        :param collection_name: target collection name.
        :type collection_name: str

        :return:
            Status: Whether the operation is successful.

        ：:rtype: Status
        """
        check_pass_param(collection_name=collection_name)

        with self._connection() as handler:
            return handler.drop_index(collection_name, timeout)

    @check_connect
    def create_partition(self, collection_name, partition_tag, timeout=30):
        """
        create a partition for a collection. 

        :param collection_name: Name of the collection.
        :type  collection_name: str

        :param partition_name: Name of the partition.
        :type  partition_name: str

        :param partition_tag: Name of the partition tag.
        :type  partition_tag: str

        :param timeout: time waiting for response.
        :type  timeout: int

        :return:
            Status: Whether the operation is successful.

        """
        check_pass_param(collection_name=collection_name, partition_tag=partition_tag)
        with self._connection() as handler:
            return handler.create_partition(collection_name, partition_tag, timeout)

    @check_connect
    def has_partition(self, collection_name, partition_tag):
        """
        Check if specified partition exists.

        :param collection_name: target table name.
        :type  collection_name: str

        :param partition_tag: partition tag.
        :type  partition_tag: str

        :return:
            Status: Whether the operation is successful.
            exists: If specified partition exists

        """
        check_pass_param(collection_name=collection_name, partition_tag=partition_tag)
        with self._connection() as handler:
            return handler.has_partition(collection_name, partition_tag)

    @check_connect
    def list_partitions(self, collection_name, timeout=30):
        """
        Show all partitions in a collection.

        :param collection_name: target table name.
        :type  collection_name: str

        :param timeout: time waiting for response.
        :type  timeout: int

        :return:
            Status: Whether the operation is successful.
            partition_list:

        """
        check_pass_param(collection_name=collection_name)

        with self._connection() as handler:
            return handler.show_partitions(collection_name, timeout)

    @check_connect
    def drop_partition(self, collection_name, partition_tag, timeout=30):
        """
        Deletes a partition in a collection.

        :param collection_name: Collection name.
        :type  collection_name: str

        :param partition_tag: Partition name.
        :type  partition_tag: str

        :param timeout: time waiting for response.
        :type  timeout: int

        :return:
            Status: Whether the operation is successful.

        """
        check_pass_param(collection_name=collection_name, partition_tag=partition_tag)
        with self._connection() as handler:
            return handler.drop_partition(collection_name, partition_tag, timeout)

    @check_connect
    def search(self, collection_name, top_k, query_records, partition_tags=None, params=None, timeout=None, **kwargs):
        """
        Search vectors in a collection.

        :param collection_name: Name of the collection.
        :type  collection_name: str

        :param top_k: number of vertors which is most similar with query vectors
        :type  top_k: int

        :param nprobe: cell number of probe
        :type  nprobe: int

        :param query_records: vectors to query
        :type  query_records: list[list[float32]]

        :param partition_tags: tags to search
        :type  partition_tags: list

        :return
            Status: Whether the operation is successful.
            result: query result

        :rtype: (Status, TopKQueryResult)

        """
        check_pass_param(collection_name=collection_name, topk=top_k, records=query_records)
        if partition_tags is not None:
            check_pass_param(partition_tag_array=partition_tags)

        params = dict() if params is None else params
        if not isinstance(params, dict):
            raise ParamError("Params must be a dictionary type")
        with self._connection() as handler:
            return handler.search(collection_name, top_k, query_records, partition_tags, params, timeout, **kwargs)

    @check_connect
    def search_hybrid_pb(self, collection_name, query_entities, partition_tags=None, params=None, **kwargs):
        with self._connection() as handler:
            return handler.search_hybrid_pb(collection_name, query_entities, partition_tags, params, **kwargs)

    @check_connect
    def search_hybrid(self, collection_name, vector_params, dsl, partition_tags=None, params=None, **kwargs):
        with self._connection() as handler:
            return handler.search_hybrid(collection_name, vector_params, dsl, partition_tags, params, **kwargs)

    @check_connect
    def search_in_segment(self, collection_name, file_ids, query_records, top_k, params=None, timeout=None, **kwargs):
        """
        Searches for vectors in specific segments of a collection.

        The Milvus server stores vector data into multiple files. Searching for vectors in specific files is a
        method used in Mishards. Obtain more detail about Mishards, see
        <a href="https://github.com/milvus-io/milvus/tree/master/shards">

        :type  collection_name: str
        :param collection_name: table name been queried

        :type  file_ids: list[str] or list[int]
        :param file_ids: Specified files id array

        :type  query_records: list[list[float]]
        :param query_records: all vectors going to be queried

        :param query_ranges: Optional ranges for conditional search.

            If not specified, search in the whole table

        :type  top_k: int
        :param top_k: how many similar vectors will be searched

        :returns:
            Status:  indicate if query is successful
            results: query result

        :rtype: (Status, TopKQueryResult)
        """
        check_pass_param(collection_name=collection_name, topk=top_k, records=query_records, ids=file_ids)

        params = dict() if params is None else params
        if not isinstance(params, dict):
            raise ParamError("Params must be a dictionary type")
        with self._connection() as handler:
            return handler.search_in_files(collection_name, file_ids,
                                           query_records, top_k, params, timeout, **kwargs)

    @check_connect
    def delete_entity_by_id(self, collection_name, id_array, timeout=None):
        """
        Deletes vectors in a collection by vector ID.

        :param collection_name: Name of the collection.
        :type  collection_name: str

        :param id_array: list of vector id
        :type  id_array: list[int]

        :return:
            Status: Whether the operation is successful.
        """
        check_pass_param(collection_name=collection_name, ids=id_array)
        with self._connection() as handler:
            return handler.delete_by_id(collection_name, id_array, timeout)

    @check_connect
    def flush(self, collection_name_array=None, timeout=None, **kwargs):
        """
        Flushes vector data in one collection or multiple collections to disk.

        :type  collection_name_array: list
        :param collection_name: Name of one or multiple collections to flush.

        """

        if collection_name_array in (None, []):
            with self._connection() as handler:
                return handler.flush([], timeout)

        if not isinstance(collection_name_array, list):
            raise ParamError("Collection name array must be type of list")

        if len(collection_name_array) <= 0:
            raise ParamError("Collection name array is not allowed to be empty")

        for name in collection_name_array:
            check_pass_param(collection_name=name)
        with self._connection() as handler:
            return handler.flush(collection_name_array, timeout, **kwargs)

    @check_connect
    def compact(self, collection_name, timeout=None, **kwargs):
        """
        Compacts segments in a collection. This function is recommended after deleting vectors.

        :type  collection_name: str
        :param collection_name: Name of the collections to compact.

        """
        check_pass_param(collection_name=collection_name)
        with self._connection() as handler:
            return handler.compact(collection_name, timeout, **kwargs)

    def get_config(self, parent_key, child_key):
        """
        Gets Milvus configurations.

        """
        cmd = "get_config {}.{}".format(parent_key, child_key)

        return self._cmd(cmd)

    def set_config(self, parent_key, child_key, value):
        """
        Sets Milvus configurations.

        """
        cmd = "set_config {}.{} {}".format(parent_key, child_key, value)

        return self._cmd(cmd)




import ujson

from ..client.exceptions import ParamError

from .check import check_pass_param

from .types import IndexType


class LoopBase(object):
    def __init__(self):
        self.__index = 0

    def __iter__(self):
        return self

    def __next__(self):
        while self.__index < self.__len__():
            self.__index += 1
            return self.__getitem__(self.__index - 1)

        # iterate stop, raise Exception
        self.__index = 0
        raise StopIteration()


class CollectionSchema:
    def __init__(self, collection_name, dimension, index_file_size, metric_type):
        """
        Table Schema

        :type  table_name: str
        :param table_name: (Required) name of table

            `IndexType`: 0-invalid, 1-flat, 2-ivflat, 3-IVF_SQ8, 4-MIX_NSG

        :type  dimension: int64
        :param dimension: (Required) dimension of vector

        :type  index_file_size: int64
        :param index_file_size: (Optional) max size of files which store index

        :type  metric_type: MetricType
        :param metric_type: (Optional) vectors metric type

            `MetricType`: 1-L2, 2-IP

        """
        check_pass_param(collection_name=collection_name, dimension=dimension,
                         index_file_size=index_file_size, metric_type=metric_type)

        self.collection_name = collection_name
        self.dimension = dimension
        self.index_file_size = index_file_size
        self.metric_type = metric_type

    def __repr__(self):
        attr_list = ['%s=%r' % (key, value) for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attr_list))


class QueryResult:

    def __init__(self, _id, _distance):
        self.id = _id
        self.distance = _distance

    def __str__(self):
        return "Result(id={}, distance={})".format(self.id, self.distance)


class RowQueryResult:
    def __init__(self, _id_list, _dis_list):
        self._id_list = _id_list or []
        self._dis_list = _dis_list or []

        # Iterator index
        self.__index = 0

    def __getitem__(self, item):
        if isinstance(item, slice):
            _start = item.start or 0
            _end = min(item.stop, self.__len__()) if item.stop else self.__len__()
            _step = item.step or 1

            elements = []
            for i in range(_start, _end, _step):
                elements.append(self.__getitem__(i))
            return elements

        return QueryResult(self._id_list[item], self._dis_list[item])

    def __len__(self):
        return len(self._id_list)

    def __iter__(self):
        return self

    def __next__(self):
        while self.__index < self.__len__():
            self.__index += 1
            return self.__getitem__(self.__index - 1)

        # iterate stop, raise Exception
        self.__index = 0
        raise StopIteration()


class TopKQueryResult:
    """
    TopK query results, shown as 2-D array

    This Class unpack response from server, store ids and distances separately.
    """

    def __init__(self, raw_source, **kwargs):
        self._raw = raw_source
        self._nq = 0
        self._topk = 0
        self._id_array = []
        self._dis_array = []

        ##
        self.__index = 0

        self._unpack(self._raw)

    def _unpack(self, _raw):
        """

        Args:
            _raw:

        Returns:

        """
        self._nq = _raw.row_num

        if self._nq == 0:
            return

        self._id_array = [list() for _ in range(self._nq)]
        self._dis_array = [list() for _ in range(self._nq)]

        id_list = list(_raw.ids)
        dis_list = list(_raw.distances)
        if len(id_list) != len(dis_list):
            raise ParamError("number of id not match distance ")

        len_ = len(id_list)
        col = len_ // self._nq if self._nq > 0 else 0

        if col == 0:
            return

        if len_ % col != 0:
            raise ValueError("Search result is not aligned. (row_num={}, len of ids={})".format(_raw.row_num, len_))

        for si, i in enumerate(range(0, len_, col)):
            k = min(i + col, len_)
            for j in range(i, k):
                if id_list[j] == -1:
                    k = j
                    break
            self._id_array[si].extend(id_list[i: k])
            self._dis_array[si].extend(dis_list[i: k])

        if len(self._id_array) != self._nq or \
                len(self._dis_array) != self._nq:
            raise ParamError("Result parse error.")

        self._topk = col

    @property
    def id_array(self):
        """
        Id array, it's a 2-D array.
        """
        return self._id_array

    @property
    def distance_array(self):
        """
        Distance array, it's a 2-D array
        """
        return self._dis_array

    @property
    def shape(self):
        """
        getter. return result shape, format as (row, column).

        """
        return self._nq, self._topk

    @property
    def raw(self):
        """
        getter. return the raw result response

        """
        return self._raw

    def __len__(self):
        return self._nq

    def __getitem__(self, item):
        if isinstance(item, slice):
            _start = item.start or 0
            _end = min(item.stop, self.__len__()) if item.stop else self.__len__()
            _step = item.step or 1

            elements = []
            for i in range(_start, _end, _step):
                elements.append(self.__getitem__(i))
            return elements

        return RowQueryResult(self._id_array[item], self._dis_array[item])

    def __iter__(self):
        return self

    def __next__(self):
        while self.__index < self.__len__():
            self.__index += 1
            return self.__getitem__(self.__index - 1)

        # iterate stop, raise Exception
        self.__index = 0
        raise StopIteration()

    def __repr__(self):
        """
        :return:
        """

        lam = lambda x: "(id:{}, distance:{})".format(x.id, x.distance)

        if self.__len__() > 5:
            middle = ''

            ll = self[:3]
            for topk in ll:
                if len(topk) > 5:
                    middle = middle + " [ %s" % ",\n   ".join(map(lam, topk[:3]))
                    middle += ",\n   ..."
                    middle += "\n   %s ]\n\n" % lam(topk[-1])
                else:
                    middle = middle + " [ %s ] \n" % ",\n   ".join(map(lam, topk))

            spaces = """        ......
            ......"""

            ahead = "[\n%s%s\n]" % (middle, spaces)
            return ahead

        # self.__len__() < 5
        str_out_list = []
        for i in range(self.__len__()):
            str_out_list.append("[\n%s\n]" % ",\n".join(map(lam, self[i])))

        return "[\n%s\n]" % ",\n".join(str_out_list)


class TopKQueryResult2:
    def __init__(self, raw_source, **kwargs):
        self._raw = raw_source
        self._nq = 0
        self._topk = 0
        self._results = []

        self.__index = 0

        self._unpack(self._raw)

    def _unpack(self, raw_resources):
        js = raw_resources.json()
        self._nq = js["num"]

        for row_result in js["result"]:
            row_ = [QueryResult(int(result["id"]), float(result["distance"]))
                    for result in row_result if float(result["id"]) != -1]

            self._results.append(row_)

    @property
    def shape(self):
        return len(self._results), len(self._results[0]) if len(self._results) > 0 else 0

    def __len__(self):
        return len(self._results)

    def __getitem__(self, item):
        return self._results.__getitem__(item)

    def __iter__(self):
        return self

    def __next__(self):
        while self.__index < self.__len__():
            self.__index += 1
            return self.__getitem__(self.__index - 1)

        self.__index = 0
        raise StopIteration()

    def __repr__(self):
        lam = lambda x: "(id:{}, distance:{})".format(x.id, x.distance)

        if self.__len__() > 5:
            middle = ''

            ll = self[:3]
            for topk in ll:
                if len(topk) > 5:
                    middle = middle + " [ %s" % ",\n   ".join(map(lam, topk[:3]))
                    middle += ",\n   ..."
                    middle += "\n   %s ]\n\n" % lam(topk[-1])
                else:
                    middle = middle + " [ %s ] \n" % ",\n   ".join(map(lam, topk))

            spaces = """        ......
                    ......"""

            ahead = "[\n%s%s\n]" % (middle, spaces)
            return ahead

        # self.__len__() < 5
        str_out_list = []
        for i in range(self.__len__()):
            str_out_list.append("[\n%s\n]" % ",\n".join(map(lam, self[i])))

        return "[\n%s\n]" % ",\n".join(str_out_list)


class HybridEntityResult:
    def __init__(self, filed_names, entity_id, entity, vector, distance):
        self._field_names = filed_names
        self._id = entity_id
        self._entity = entity
        self._vector = vector
        self._distance = distance

    def __getattr__(self, item):
        index = self._field_names.index(item)
        if index < len(self._entity):
            return self._entity[index]
        if index == len(self._entity):
            return list(self._vector.float_data) if len(self._vector.float_data) > 0 else bytes(
                self._vector.binary_data)

        raise ValueError("Out of range ... ")

    # def __getitem__(self, item):
    #     return self._entity.__getitem__(item)

    def get(self, field):
        return self.__getattr__(field)

    @property
    def id(self):
        return self._id

    @property
    def distance(self):
        return self._distance


class HybridRawResult(LoopBase):
    def __init__(self, field_names, ids, entities, vectors, distances):
        super().__init__()
        self._field_names = field_names
        self._ids = ids
        self._vectors = vectors
        self._entities = entities
        self._distances = distances

    def __len__(self):
        return len(self._ids)

    def __getitem__(self, item):
        if isinstance(item, slice):
            _start = item.start or 0
            _end = min(item.stop, self.__len__()) if item.stop else self.__len__()
            _step = item.step or 1

            elements = []
            for i in range(_start, _end, _step):
                elements.append(self.__getitem__(i))
            return elements

        item_entities = [e[item] for e in self._entities]

        return HybridEntityResult(self._field_names, self._ids[item], item_entities, self._vectors[item],
                                  self._distances[item])
        # if item in self._field_names:
        #     index = self._field_names[item]


class HybridResult(LoopBase):

    def __init__(self, raw, **kwargs):
        super().__init__()
        self._raw = raw

    def __len__(self):
        return self._raw.row_num

    def __getitem__(self, item):
        if isinstance(item, slice):
            _start = item.start or 0
            _end = min(item.stop, self.__len__()) if item.stop else self.__len__()
            _step = item.step or 1

            elements = []
            for i in range(_start, _end, _step):
                elements.append(self.__getitem__(i))
            return elements

        seg_size = len(self._raw.distance) // self.row_num()
        seg_start = item * seg_size
        seg_end = min((item + 1) * seg_size, len(self._raw.distance))

        entities = self._raw.entity

        slice_entity = lambda e, start, end: e.int_value[start: end] if len(e.int_value) > 0 else e.double_value[
                                                                                                  start: end]
        seg_ids = self._raw.entity.entity_id[seg_start: seg_end]
        seg_entities = [slice_entity(e, seg_start, seg_end) for e in entities.attr_data]
        seg_vectors = self._raw.entity.vector_data[0].value[seg_start: seg_end]
        seg_distances = self._raw.distance[seg_start: seg_end]
        return HybridRawResult(self.field_names(), seg_ids, seg_entities, seg_vectors, seg_distances)

    def field_names(self):
        return list(self._raw.entity.field_names)

    def row_num(self):
        return self._raw.row_num


class HEntity:
    def __init__(self, field_names, attr_records, vector):
        self._field_names = field_names
        self._attr_records = attr_records
        self._vector = vector

    def get(self, field):
        index = self._field_names.index(field)
        if index < len(self._attr_records):
            return self._attr_records[index]

        if index == len(self._attr_records):
            return self._vector

        raise ValueError("Out of range ... ")


class HEntitySet(LoopBase):
    def __init__(self, raw):
        super().__init__()
        self._raw = raw

    def __len__(self):
        return self._raw.row_num

    def __getitem__(self, item):
        if isinstance(item, slice):
            _start = item.start or 0
            _end = min(item.stop, self.__len__()) if item.stop else self.__len__()
            _step = item.step or 1

            elements = []
            for i in range(_start, _end, _step):
                elements.append(self.__getitem__(i))
            return elements

        slice_entity = lambda e, index: e.int_value[index] if len(e.int_value) > 0 else e.double_value[index]
        attr_records = [slice_entity(e, item) for e in self._raw.attr_data]
        vector = self._raw.vector_data[0].value[item]

        return HEntity(self.field_names, attr_records, vector)

    @property
    def field_names(self):
        return list(self._raw.field_names)


class IndexParam:
    """
    Index Param

    :type  table_name: str
    :param table_name: (Required) name of table

    :type  index_type: IndexType
    :param index_type: (Required) index type, default = IndexType.INVALID

        `IndexType`: 0-invalid, 1-flat, 2-ivflat, 3-IVF_SQ8, 4-MIX_NSG

    :type  nlist: int64
    :param nlist: (Required) num of cell

    """

    def __init__(self, collection_name, index_type, params):

        if collection_name is None:
            raise ParamError('Collection name can\'t be None')
        collection_name = str(collection_name) if not isinstance(collection_name, str) else collection_name

        if isinstance(index_type, int):
            index_type = IndexType(index_type)
        if not isinstance(index_type, IndexType) or index_type == IndexType.INVALID:
            raise ParamError('Illegal index_type, should be IndexType but not IndexType.INVALID')

        self._collection_name = collection_name
        self._index_type = index_type

        if not isinstance(params, dict):
            self._params = ujson.loads(params)
        else:
            self._params = params

    def __str__(self):
        attr_list = ['%s=%r' % (key.lstrip('_'), value)
                     for key, value in self.__dict__.items()]
        return '(%s)' % (', '.join(attr_list))

    def __repr__(self):
        attr_list = ['%s=%r' % (key, value)
                     for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attr_list))

    @property
    def collection_name(self):
        return self._collection_name

    @property
    def index_type(self):
        return self._index_type

    @property
    def params(self):
        return self._params


class PartitionParam:

    def __init__(self, collection_name, tag):
        self.collection_name = collection_name
        self.tag = tag

    def __repr__(self):
        attr_list = ['%s=%r' % (key, value)
                     for key, value in self.__dict__.items()]
        return '(%s)' % (', '.join(attr_list))


class SegmentStat:
    def __init__(self, res):
        self.segment_name = res.segment_name
        self.count = res.row_count
        self.index_name = res.index_name
        self.data_size = res.data_size

    def __str__(self):
        attr_list = ['%s: %r' % (key, value)
                     for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attr_list))

    def __repr__(self):
        return self.__str__()


class PartitionStat:
    def __init__(self, res):
        self.tag = res.tag
        self.count = res.total_row_count
        self.segments_stat = [SegmentStat(s) for s in list(res.segments_stat)]

    def __str__(self):
        attr_list = ['%s: %r' % (key, value)
                     for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attr_list))

    def __repr__(self):
        return self.__str__()


class CollectionInfo:
    def __init__(self, res):
        self.count = res.total_row_count
        self.partitions_stat = [PartitionStat(p) for p in list(res.partitions_stat)]

    def __str__(self):
        attr_list = ['%s: %r' % (key, value)
                     for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attr_list))

    def __repr__(self):
        return self.__str__()


class HSegmentStat:
    def __init__(self, res):
        self.segment_name = res["segment_name"]
        self.count = res["count"]
        self.index_name = res["index"]
        self.data_size = res["size"]

    def __str__(self):
        attr_list = ['%s: %r' % (key, value)
                     for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attr_list))

    def __repr__(self):
        return self.__str__()


class HPartitionStat:
    def __init__(self, res):
        self.tag = res["partition_tag"]
        self.count = res["count"]
        self.segments_stat = [HSegmentStat(s) for s in res["segments_stat"]]

    def __str__(self):
        attr_list = ['%s: %r' % (key, value)
                     for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attr_list))

    def __repr__(self):
        return self.__str__()


class HCollectionInfo:
    def __init__(self, res):
        self.count = res["count"]
        self.partitions_stat = [HPartitionStat(p) for p in res["partitions_stat"]]

    def __str__(self):
        attr_list = ['%s: %r' % (key, value)
                     for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(attr_list))

    def __repr__(self):
        return self.__str__()


def _abstract():
    raise NotImplementedError('You need to override this function')


class ConnectIntf:
    """SDK client abstract class

    Connection is a abstract class

    """

    def connect(self, host, port, uri, timeout):
        """
        Connect method should be called before any operations
        Server will be connected after connect return OK
        Should be implemented

        :type  host: str
        :param host: host

        :type  port: str
        :param port: port

        :type  uri: str
        :param uri: (Optional) uri

        :type  timeout: int
        :param timeout:

        :return: Status,  indicate if connect is successful
        """
        _abstract()

    def connected(self):
        """
        connected, connection status
        Should be implemented

        :return: Status,  indicate if connect is successful
        """
        _abstract()

    def disconnect(self):
        """
        Disconnect, server will be disconnected after disconnect return SUCCESS
        Should be implemented

        :return: Status,  indicate if connect is successful
        """
        _abstract()

    def create_table(self, param, timeout):
        """
        Create table
        Should be implemented

        :type  param: TableSchema
        :param param: provide table information to be created

        :type  timeout: int
        :param timeout:

        :return: Status, indicate if connect is successful
        """
        _abstract()

    def has_table(self, table_name, timeout):
        """

        This method is used to test table existence.
        Should be implemented

        :type table_name: str
        :param table_name: table name is going to be tested.

        :type  timeout: int
        :param timeout:

        :return:
            has_table: bool, if given table_name exists

        """
        _abstract()

    def delete_table(self, table_name, timeout):
        """
        Delete table
        Should be implemented

        :type  table_name: str
        :param table_name: table_name of the deleting table

        :type  timeout: int
        :param timeout:

        :return: Status, indicate if connect is successful
        """
        _abstract()

    def add_vectors(self, table_name, records, ids, timeout, **kwargs):
        """
        Add vectors to table
        Should be implemented

        :type  table_name: str
        :param table_name: table name been inserted

        :type  records: list[RowRecord]
        :param records: list of vectors been inserted

        :type  ids: list[int]
        :param ids: list of ids

        :type  timeout: int
        :param timeout:

        :returns:
            Status : indicate if vectors inserted successfully
            ids :list of id, after inserted every vector is given a id
        """
        _abstract()

    def search_vectors(self, table_name, top_k, nprobe, query_records, query_ranges, **kwargs):
        """
        Query vectors in a table
        Should be implemented

        :type  table_name: str
        :param table_name: table name been queried

        :type  query_records: list[RowRecord]
        :param query_records: all vectors going to be queried

        :type  query_ranges: list[Range]
        :param query_ranges: Optional ranges for conditional search.
            If not specified, search whole table

        :type  top_k: int
        :param top_k: how many similar vectors will be searched

        :returns:
            Status:  indicate if query is successful
            query_results: list[TopKQueryResult]
        """
        _abstract()

    def search_vectors_in_files(self, table_name, file_ids, query_records,
                                top_k, nprobe, query_ranges, **kwargs):
        """
        Query vectors in a table, query vector in specified files
        Should be implemented

        :type  table_name: str
        :param table_name: table name been queried

        :type  file_ids: list[str]
        :param file_ids: Specified files id array

        :type  query_records: list[RowRecord]
        :param query_records: all vectors going to be queried

        :type  query_ranges: list[Range]
        :param query_ranges: Optional ranges for conditional search.
            If not specified, search whole table

        :type  top_k: int
        :param top_k: how many similar vectors will be searched

        :returns:
            Status:  indicate if query is successful
            query_results: list[TopKQueryResult]
        """
        _abstract()

    def describe_table(self, table_name, timeout):
        """
        Show table information
        Should be implemented

        :type  table_name: str
        :param table_name: which table to be shown

        :type  timeout: int
        :param timeout:

        :returns:
            Status: indicate if query is successful
            table_schema: TableSchema, given when operation is successful
        """
        _abstract()

    def get_table_row_count(self, table_name, timeout):
        """
        Get table row count
        Should be implemented

        :type  table_name, str
        :param table_name, target table name.

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :returns:
            Status: indicate if operation is successful
            count: int, table row count
        """
        _abstract()

    def show_tables(self, timeout):
        """
        Show all tables in database
        should be implemented

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if this operation is successful
            tables: list[str], list of table names
        """
        _abstract()

    def create_index(self, table_name, index, timeout):
        """
        Create specified index in a table
        should be implemented

        :type  table_name: str
        :param table_name: table name

         :type index: dict
        :param index: index information dict

            example: index = {
                "index_type": IndexType.FLAT,
                "nlist": 18384
            }

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if this operation is successful

        :rtype: Status
        """
        _abstract()

    def server_version(self, timeout):
        """
        Provide server version
        should be implemented

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

            str : Server version

        :rtype: (Status, str)
        """
        _abstract()

    def server_status(self, timeout):
        """
        Provide server status. When cmd !='version', provide 'OK'
        should be implemented

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

            str : Server version

        :rtype: (Status, str)
        """
        _abstract()

    def preload_table(self, table_name, timeout):
        """
        load table to memory cache in advance
        should be implemented

        :param table_name: target table name.
        :type table_name: str

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

        ：:rtype: Status
        """

        _abstract()

    def describe_index(self, table_name, timeout):
        """
        Show index information
        should be implemented

        :param table_name: target table name.
        :type table_name: str

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

            TableSchema: table detail information

        :rtype: (Status, TableSchema)
        """

        _abstract()

    def drop_index(self, table_name, timeout):
        """
        Show index information
        should be implemented

        :param table_name: target table name.
        :type table_name: str

        :type  timeout: int
        :param timeout: how many similar vectors will be searched

        :return:
            Status: indicate if operation is successful

        ：:rtype: Status
        """

        _abstract()



import copy
import struct
import ujson

from .exceptions import ParamError

from ..grpc_gen import milvus_pb2 as grpc_types
from ..grpc_gen import status_pb2

from .types import RangeType

BoolOccurMap = {
    "must": grpc_types.MUST,
    "must_not": grpc_types.MUST_NOT,
    "should": grpc_types.SHOULD
}

RangeOperatorMap = {
    RangeType.LT: grpc_types.LT,
    RangeType.LTE: grpc_types.LTE,
    RangeType.EQ: grpc_types.EQ,
    RangeType.GT: grpc_types.GT,
    RangeType.GTE: grpc_types.GTE,
    RangeType.NE: grpc_types.NE
}


class Prepare:

    @classmethod
    def collection_name(cls, collection_name):

        return grpc_types.CollectionName(collection_name=collection_name)

    @classmethod
    def collection_schema(cls, collection_name, dimension, index_file_size, metric_type, param):
        """
        :type param: dict
        :param param: (Required)

            `example param={'collection_name': 'name',
                            'dimension': 16,
                            'index_file_size': 1024
                            'metric_type': MetricType.L2
                            }`

        :return: ttypes.TableSchema object
        """

        _param = grpc_types.CollectionSchema(status=status_pb2.Status(error_code=0, reason='Client'),
                                             collection_name=collection_name,
                                             dimension=dimension,
                                             index_file_size=index_file_size,
                                             metric_type=metric_type)

        if param:
            param_str = ujson.dumps(param)
            _param.extra_params.add(key="params", value=param_str)

        return _param

    @classmethod
    def collection_hybrid_schema(cls, collection_name, fields):
        _param = grpc_types.Mapping(
            collection_name=collection_name
        )

        for field in fields:
            if "data_type" in field:
                ft = grpc_types.FieldType(data_type=int(field["data_type"]))
                # ft.data_type = int(v["data_type"])
            elif "dimension" in field:
                ft = grpc_types.FieldType(vector_param=grpc_types.VectorFieldParam(dimension=field["dimension"]))
                # ft.vector_param = grpc_types.VectorFieldParam(dimension=v["dimension"])
            else:
                raise ValueError("Collection field not support {}".format(field))
            field_param = grpc_types.FieldParam(name=field["field_name"], type=ft)
            extra_params = field.get("extra_params", None)
            if extra_params:
                u = ujson.dumps(extra_params)
                field_param.extra_params.add(key="params", value=u)
            _param.fields.append(field_param)

        return _param

    @classmethod
    def preload_param(cls, collection_name, partition_tags):
        return grpc_types.PreloadCollectionParam(collection_name=collection_name, partition_tag_array=partition_tags)

    @classmethod
    def reload_param(cls, collection_name, segment_ids):
        return grpc_types.ReLoadSegmentsParam(collection_name=collection_name, segment_id_array=segment_ids)

    @classmethod
    def insert_param(cls, collection_name, vectors, partition_tag, ids=None, params=None, **kwargs):
        if ids is None:
            _param = grpc_types.InsertParam(collection_name=collection_name, partition_tag=partition_tag)
        else:
            _param = grpc_types.InsertParam(
                collection_name=collection_name,
                row_id_array=ids,
                partition_tag=partition_tag)

        for vector in vectors:
            if isinstance(vector, bytes):
                _param.row_record_array.add(binary_data=vector)
            else:
                _param.row_record_array.add(float_data=vector)

        params = params or dict()
        params_str = ujson.dumps(params)
        _param.extra_params.add(key="params", value=params_str)

        return _param

    @classmethod
    def insert_hybrid_param(cls, collection_name, tag, entities, vector_entities, ids=None, params=None):
        entity_param = grpc_types.HEntity()

        _len = -1
        for entity in entities:
            values = entity["field_values"]
            if not isinstance(values, list):
                raise ValueError("Field values must be a list")
            if _len == -1:
                _len = len(values)
            else:
                if len(values) != _len:
                    raise ValueError("Length is not equal")

        for entity in entities:
            entity_param.field_names.append(entity["field_name"])
            values = entity["field_values"]
            if isinstance(values, list):
                if isinstance(values[0], int):
                    entity_param.attr_data.add(int_value=values)
                elif isinstance(values[0], float):
                    entity_param.attr_data.add(double_value=values)
                else:
                    raise ValueError("Field item must be int or float")
            else:
                raise ValueError("Field values must be a list")
        # entity_param.attr_records = bytes(item_bytes)
        entity_param.row_num = _len
        # vectors
        # entity.field_names.append(vector_field)
        for vector_entity in vector_entities:
            entity_param.field_names.append(vector_entity["field_name"])
            vector_field = grpc_types.VectorFieldRecord()
            vectors = vector_entity["field_values"]
            for vector in vectors:
                if isinstance(vector, bytes):
                    vector_field.value.add(binary_data=vector)
                else:
                    vector_field.value.add(float_data=vector)
            entity_param.vector_data.append(vector_field)

        h_param = grpc_types.HInsertParam(
            collection_name=collection_name,
            partition_tag=tag,
            entity=entity_param
        )

        if ids:
            h_param.entity_id_array[:] = ids
        params = params or dict()
        params_str = ujson.dumps(params)
        h_param.extra_params.add(key="params", value=params_str)
        return h_param

    @classmethod
    def index_param(cls, collection_name, index_type, params):

        _param = grpc_types.IndexParam(status=status_pb2.Status(error_code=0, reason='Client'),
                                       collection_name=collection_name,
                                       index_type=index_type)
        params = params or dict()
        params_str = ujson.dumps(params)
        _param.extra_params.add(key="params", value=params_str)

        return _param

    @classmethod
    def search_param(cls, collection_name, topk, query_records, partitions, params):

        search_param = grpc_types.SearchParam(
            collection_name=collection_name,
            topk=topk,
            partition_tag_array=partitions
        )

        for vector in query_records:
            if isinstance(vector, bytes):
                search_param.query_record_array.add(binary_data=vector)
            else:
                search_param.query_record_array.add(float_data=vector)

        params = params or dict()
        params_str = ujson.dumps(params)
        search_param.extra_params.add(key="params", value=params_str)

        return search_param

    @classmethod
    def search_hybrid_pb_param(cls, collection_name, query_entities, partition_tags, params):

        def term_query(node):
            if len(node) > 1:
                raise Exception()
            for k, v in node.items():
                vs = v.get("values", None)
                if not vs:
                    raise ValueError("Key values is missing")

                _term_param = grpc_types.TermQuery(field_name=k,
                                                   value_num=len(v["values"]),
                                                   # boost=node["boost"]
                                                   )
                if isinstance(vs, list):
                    if isinstance(vs[0], int):
                        _term_param.int_value[:] = vs
                    elif isinstance(vs[0], float):
                        _term_param.double_value[:] = vs
                    else:
                        raise ValueError("Field item must be int or float")
                else:
                    raise ValueError("Field values must be a list")
                return _term_param

        def range_query(node):
            if len(node) > 1:
                raise Exception("Item size > 1")
            for name, query in node.items():
                _range_param = grpc_types.RangeQuery(field_name=name,
                                                     # boost=node["boost"]
                                                     )
                for k, v in query["ranges"].items():
                    ope = RangeOperatorMap[k]
                    _range_param.operand.add(operator=ope, operand=str(v))

                return _range_param

        def vector_query(node):
            if len(node) > 1:
                raise Exception("Item size > 1")
            for name, query in node.items():
                _vector_param = grpc_types.VectorQuery(field_name=name,
                                                       # query_boost=node["boost"],
                                                       topk=query["topk"]
                                                       )
                for vector in query["query"]:
                    if isinstance(vector, bytes):
                        _vector_param.records.add(binary_data=vector)
                    else:
                        _vector_param.records.add(float_data=vector)

                _extra_param = query.get("params", None)

                _extra_param = _extra_param or dict()
                params_str = ujson.dumps(_extra_param)
                _vector_param.extra_params.add(key="params", value=params_str)
                return _vector_param

        def gene_node(key, node):
            if isinstance(node, list):
                bqr = grpc_types.BooleanQuery(occur=BoolOccurMap[key])
                for query in node:
                    if "term" in query:
                        # bqr.general_query.append(grpc_types.GeneralQuery(term_query=term_query(query["term"])))
                        bqr.general_query.add(term_query=term_query(query["term"]))
                    elif "range" in query:
                        # bqr.general_query.append(grpc_types.GeneralQuery(range_query=range_query(query["range"])))
                        bqr.general_query.add(range_query=range_query(query["range"]))
                    elif "vector" in query:
                        # bqr.general_query.append(grpc_types.GeneralQuery(vector_query=vector_query(query["vector"])))
                        bqr.general_query.add(vector_query=vector_query(query["vector"]))
                    else:
                        raise ValueError("Unknown ")

                return grpc_types.GeneralQuery(boolean_query=bqr)

            keys = node.keys()
            sq = {"must", "must_not", "should"}
            if len(keys) + len(sq) > len(set(keys) | sq):
                gqs = list()
                for k, v in node.items():
                    gq = gene_node(k, v)
                    gqs.append(gq)
                if len(gqs) == 1:
                    return gqs[0]

                bq0 = grpc_types.BooleanQuery(occur=grpc_types.INVALID)
                for g in gqs:
                    bq0.general_query.append(g)
                return grpc_types.GeneralQuery(boolean_query=bq0)

            # bqr = grpc_types.BooleanQuery(occur=BoolOccurMap[key])
            # for k, v in node.items():
            #     field_name = node["field_name"]
            #     if k == "term":
            #         bqr.general_query.append(grpc_types.GeneralQuery(term_query=term_query(v)))
            #     elif k == "range":
            #         bqr.general_query.append(grpc_types.GeneralQuery(range_query=range_query(v)))
            #     elif k == "vector":
            #         bqr.general_query.append(grpc_types.GeneralQuery(vector_query=vector_query(v)))
            #     else:
            #         raise ValueError("Unknown ")
            #
            # return grpc_types.GeneralQuery(boolean_query=bqr)

            # if len(node) == 1:
            #     for k, v in node.items():
            #         if k in ("must", "must_not", "should"):
            #             bq = grpc_types.BooleanQuery(occur=BoolOccurMap[k])
            #
            # for k, v in node.items():
            #     if k in ("must", "must_not", "should"):
            #         len(node) == 1:
            #         vqq = grpc_types.BooleanQuery(occur=grpc_types.INVALID)
            #     if k in ("must", "must_not", "should"):
            #         bq = grpc_types.BooleanQuery(occur=BoolOccurMap[k])
            #         vqq.general_query.append(grpc_types.GeneralQuery(boolean_query=bq))

        _param = grpc_types.HSearchParamPB(
            collection_name=collection_name,
            partition_tag_array=partition_tags
        )

        _param.general_query.CopyFrom(gene_node(None, query_entities["bool"]))

        for k, v in query_entities.items():
            if k == "bool":
                continue
            _param.extra_params.add(key=k, value=ujson.dumps(v))

        # import pdb;pdb.set_trace()
        # grpc_types.GeneralQuery(boolean_query=bool_node(query_entities))
        # params = params or dict()
        # params_str = ujson.dumps(params)
        # _param.extra_params.add(key="params", value=params_str)

        return _param

    @classmethod
    def search_hybrid_param(cls, collection_name, vector_params, dsl, partition_tags, params):
        # def replace_range_item(d):
        #     if not isinstance(d, dict):
        #         return
        #
        #     if "range" not in d:
        #         for ki, vi in d.itmes():
        #             replace_range_item(vi)
        #     else:
        #         range = d["range"]
        #         for ki, vi in range.itmes():
        #             ranges = vi["values"]
        #             for kii, vii in ranges.items():
        #                 ranges.pop(kii)
        #                 ranges[int(kii)] = vii
        #         return

        # dsl_out = copy.deepcopy(dsl)
        # replace_range_item(dsl_out)

        dsl_str = dsl if isinstance(dsl, str) else ujson.dumps(dsl)
        hybrid_param = grpc_types.HSearchParam(collection_name=collection_name,
                                               partition_tag_array=partition_tags or [],
                                               dsl=dsl_str)

        for v_p in vector_params:
            if "vector" not in v_p:
                raise ParamError("Vector param must contains key \'vector\'")
            # TODO: may need to copy vector_params
            query_vectors = v_p.pop("vector")
            json_ = ujson.dumps(v_p)

            vector_param = grpc_types.VectorParam(json=json_)
            for vector in query_vectors:
                if isinstance(vector, bytes):
                    vector_param.row_record.add(binary_data=vector)
                else:
                    vector_param.row_record.add(float_data=vector)
            hybrid_param.vector_param.append(vector_param)

        _params = params or dict()
        for k, v in _params.items():
            hybrid_param.extra_params.add(key=k, value=ujson.dumps(v))

        return hybrid_param

    @classmethod
    def search_by_ids_param(cls, collection_name, ids, top_k, partition_tag_array, params):
        _param = grpc_types.SearchByIDParam(
            collection_name=collection_name, id_array=ids, topk=top_k,
        )
        if partition_tag_array:
            _param.partition_tag_array[:] = partition_tag_array

        params = params or dict()
        params_str = ujson.dumps(params)
        _param.extra_params.add(key="params", value=params_str)

        return _param

    @classmethod
    def search_vector_in_files_param(cls, collection_name, query_records, topk, ids, params):
        _search_param = Prepare.search_param(collection_name, topk, query_records,
                                             partitions=None, params=params)

        return grpc_types.SearchInFilesParam(
            file_id_array=ids,
            search_param=_search_param
        )

    @classmethod
    def cmd(cls, cmd):

        return grpc_types.Command(cmd=cmd)

    @classmethod
    def partition_param(cls, collection_name, tag):

        return grpc_types.PartitionParam(collection_name=collection_name, tag=tag)

    @classmethod
    def delete_by_id_param(cls, collection_name, id_array):

        return grpc_types.DeleteByIDParam(collection_name=collection_name, id_array=id_array)

    @classmethod
    def flush_param(cls, collection_names):

        return grpc_types.FlushParam(collection_name_array=collection_names)

    @classmethod
    def compact_param(cls, collection_name):
        return grpc_types.CollectionName(collection_name=collection_name)



import copy
import functools
import json
import logging
import ujson
from urllib.parse import urlparse
import struct

import requests as rq

from .abstract import ConnectIntf, IndexParam, CollectionSchema, TopKQueryResult2, PartitionParam
from .check import is_legal_host, is_legal_port
from .exceptions import NotConnectError, ParamError
from .types import Status, IndexType, MetricType

from ..settings import DefaultConfig as config

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

IndexValue2NameMap = {
    IndexType.INVALID: "INVALID",
    IndexType.FLAT: "FLAT",
    IndexType.IVFLAT: "IVFFLAT",
    IndexType.IVF_SQ8: "IVFSQ8",
    IndexType.IVF_SQ8H: "IVFSQ8H",
    IndexType.IVF_PQ: "IVFPQ",
    IndexType.RNSG: "RNSG",
    IndexType.HNSW: "HNSW",
    IndexType.ANNOY: "ANNOY"
}

IndexName2ValueMap = {
    "INVALID": IndexType.INVALID,
    "FLAT": IndexType.FLAT,
    "IVFFLAT": IndexType.IVFLAT,
    "IVFSQ8": IndexType.IVF_SQ8,
    "IVFSQ8H": IndexType.IVF_SQ8H,
    "IVFPQ": IndexType.IVF_PQ,
    "RNSG": IndexType.RNSG,
    "HNSW": IndexType.HNSW,
    "ANNOY": IndexType.ANNOY
}

MetricValue2NameMap = {
    MetricType.L2: "L2",
    MetricType.IP: "IP",
    MetricType.HAMMING: "HAMMING",
    MetricType.JACCARD: "JACCARD",
    MetricType.TANIMOTO: "TANIMOTO",
    MetricType.SUBSTRUCTURE: "SUBSTRUCTURE",
    MetricType.SUPERSTRUCTURE: "SUPERSTRUCTURE"
}

MetricName2ValueMap = {
    "L2": MetricType.L2,
    "IP": MetricType.IP,
    "HAMMING": MetricType.HAMMING,
    "JACCARD": MetricType.JACCARD,
    "TANIMOTO": MetricType.TANIMOTO,
    "SUBSTRUCTURE": MetricType.SUBSTRUCTURE,
    "SUPERSTRUCTURE": MetricType.SUPERSTRUCTURE
}


LOGGER = logging.getLogger(__name__)


def handle_error(returns=tuple()):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            nonlocal returns
            try:
                return func(self, *args, **kwargs)
            except rq.exceptions.Timeout:
                status = Status(Status.UNEXPECTED_ERROR, message='Request timeout')
                return returns if not returns else tuple([status]) + returns
            except json.decoder.JSONDecodeError as e:
                status = Status(Status.UNEXPECTED_ERROR, message=str(e))
                return returns if not returns else tuple([status]) + returns

        return wrapper

    return decorator


class HttpHandler(ConnectIntf):

    def __init__(self, host=None, port=None, **kwargs):
        self._status = None

        _uri = kwargs.get("uri", None)

        self._uri = (host or port or _uri) and self._set_uri(host, port, uri=_uri)
        self._max_retry = kwargs.get("max_retry", 3)

    def __enter__(self):
        self.ping()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def _set_uri(self, host, port, uri):
        """
        Set server network address

        :raises: ParamError

        """
        if host is not None:
            _port = port if port is not None else config.HTTP_PORT
            _host = host
        elif port is None:
            try:
                _uri = urlparse(uri) if uri else urlparse(config.HTTP_URI)
                _host = _uri.hostname
                _port = _uri.port
            except (AttributeError, ValueError, TypeError) as e:
                raise ParamError("uri is illegal: {}".format(e))
        else:
            raise ParamError("Param is not complete. Please invoke as follow:\n"
                             "\t(host = ${HOST}, port = ${PORT})\n"
                             "\t(uri = ${URI})\n")

        if not is_legal_host(_host) or not is_legal_port(_port):
            raise ParamError("host or port is illeagl")

        return "http://{}:{}".format(str(_host), str(_port))

    @property
    def status(self):
        return self._status

    def ping(self, timeout=10):
        if self._uri is None:
            self._uri = self._set_uri(None, None, None)
        logging.info("Connecting server {}".format(self._uri))
        retry = self._max_retry
        try:
            while retry > 0:
                try:
                    response = rq.get(self._uri + "/state", timeout=timeout)
                    return True
                except:
                    retry -= 1
                    if retry > 0:
                        continue
                    else:
                        raise
        except:
            LOGGER.error("Cannot connect server {}".format(self._uri))
            raise NotConnectError("Cannot get server status")

        LOGGER.info("Connected server {}".format(self._uri))
        # try:
        #     js = response.json()
        #     return Status(js["code"], js["message"])
        # except ValueError:
        #     pass
        #
        # return Status(Status.UNEXPECTED_ERROR, "Error occurred when parse response.")

    def set_hook(self, **kwargs):
        pass

    def connect(self, host, port, uri, timeout):
        if self.connected():
            return Status(message="You have already connected {} !".format(self._uri),
                          code=Status.CONNECT_FAILED)

        if (host or port or uri) or not self._uri:
            # if self._uri:
            #     return Status(message="The server address is set as {}, "
            #                           "you cannot connect other server".format(self._uri),
            #                   code=Status.CONNECT_FAILED)
            # else:
            self._uri = self._set_uri(host, port, uri=uri)

        self._status = self.ping()
        return self._status

    def connected(self):
        return False if not (self._status and self._status.OK()) else self.ping()

    def disconnect(self):
        self._status = None
        return Status()

    def _set_config(self, cmd, timeout):
        if cmd.startswith("set_config"):
            cmd_node = cmd.split(" ")
            config_node = cmd_node[1].split(".")
            request = {
                config_node[0]: {
                    config_node[1]: cmd_node[2]
                }
            }

            url = self._uri + "/system/config"
            payload = ujson.dumps(request)
            response = rq.put(url, data=payload, timeout=timeout)
            if response.status_code == 200:
                js = response.json()
                return Status(), js["message"]
            elif response.status_code == 400:
                js = response.json()
                return Status(js["code"], js["message"]), None
            else:
                return Status(Status.UNEXPECTED_ERROR, response.reason)

    def _get_config(self, cmd, timeout):
        if cmd.startswith("get_config"):
            cmd_node = cmd.split(" ")
            config_node = cmd_node[1].split(".")

            url = self._uri + "/system/config"
            response = rq.get(url, timeout=timeout)
            if response.status_code == 200:
                js = response.json()
                rc_parent = js.get(config_node[0], None)
                if rc_parent is None:
                    return Status(Status.UNEXPECTED_ERROR, "Config {} not supported".format(cmd_node[1]))
                rc_child = rc_parent.get(config_node[1], None)
                if rc_child is None:
                    return Status(Status.UNEXPECTED_ERROR, "Config {} not supported".format(cmd_node[1]))

                return Status(), rc_child
                # return Status(), js["message"]
            elif response.status_code == 400:
                js = response.json()
                return Status(js["code"], js["message"]), None
            else:
                return Status(Status.UNEXPECTED_ERROR, response.reason)

    @handle_error(returns=(None,))
    def _cmd(self, cmd, timeout=10):
        if cmd.startswith("get_config"):
            return self._get_config(cmd, timeout)
        if cmd.startswith("set_config"):
            return self._set_config(cmd, timeout)

        url = self._uri + "/system/{}".format(cmd)

        response = rq.get(url, timeout=timeout)

        js = response.json()
        if response.status_code == 200:
            return Status(), js["reply"]

        return Status(code=js["code"], message=js["message"]), None

    def server_version(self, timeout):
        return self._cmd("version", timeout)

    def server_status(self, timeout):
        return self._cmd("status", timeout)

    @handle_error()
    def create_collection(self, collection_name, dimension, index_file_size, metric_type, params=None, timeout=10):
        metric = MetricValue2NameMap.get(metric_type, None)

        table_param = {
            "collection_name": collection_name,
            "dimension": dimension,
            "index_file_size": index_file_size,
            "metric_type": metric
        }

        data = ujson.dumps(table_param)
        url = self._uri + "/collections"

        try:
            response = rq.post(url, data=data)
            if response.status_code == 201:
                return Status(message='Create table successfully!')

            js = response.json()
            return Status(js["code"], js["message"])
        except Exception as e:
            return Status(Status.UNEXPECTED_ERROR, message=str(e))

    @handle_error(returns=(False,))
    def has_collection(self, table_name, timeout):
        url = self._uri + "/collections/" + table_name
        try:
            response = rq.get(url=url, timeout=timeout)
            if response.status_code == 200:
                return Status(), True

            if response.status_code == 404:
                return Status(), False

            js = response.json()
            return Status(js["code"], js["message"]), False
        except Exception as e:
            return Status(Status.UNEXPECTED_ERROR, message=str(e))

    @handle_error(returns=(None,))
    def count_collection(self, table_name, timeout):
        url = self._uri + "/collections/{}".format(table_name)

        try:
            response = rq.get(url, timeout=timeout)
            js = response.json()

            if response.status_code == 200:
                return Status(), js["count"]

            return Status(js["code"], js["message"]), None
        except Exception as e:
            return Status(Status.UNEXPECTED_ERROR, message=str(e)), None

    @handle_error(returns=(None,))
    def describe_collection(self, table_name, timeout):
        url = self._uri + "/collections/{}".format(table_name)

        response = rq.get(url, timeout=timeout)

        if response.status_code >= 500:
            return Status(Status.UNEXPECTED_ERROR, response.reason), None

        js = response.json()
        if response.status_code == 200:
            metric_map = dict()
            _ = [metric_map.update({i.name: i.value}) for i in MetricType if i.value > 0]

            table = CollectionSchema(
                collection_name=js["collection_name"],
                dimension=js["dimension"],
                index_file_size=js["index_file_size"],
                metric_type=metric_map[js["metric_type"]])

            return Status(message='Describe table successfully!'), table

        return Status(js["code"], js["message"]), None

    @handle_error(returns=([],))
    def show_collections(self, timeout):
        url = self._uri + "/collections"

        response = rq.get(url, params={"offset": 0, "page_size": 0}, timeout=timeout)
        if response.status_code != 200:
            return Status(Status.UNEXPECTED_ERROR, response.reason), []

        js = response.json()
        count = js["count"]

        response = rq.get(url, params={"offset": 0, "page_size": count}, timeout=timeout)
        if response.status_code != 200:
            return Status(Status.UNEXPECTED_ERROR, response.reason), []

        tables = []
        js = response.json()

        for table in js["collections"]:
            tables.append(table["collection_name"])

        return Status(), tables

    @handle_error(returns=(None,))
    def show_collection_info(self, table_name, timeout=10):
        url = self._uri + "/collections/{}?info=stat".format(table_name)

        response = rq.get(url, timeout=timeout)
        if response.status_code == 200:
            return Status(), response.json()

        if response.status_code == 404:
            return Status(Status.COLLECTION_NOT_EXISTS, "Collection not found"), None

        if response.text:
            result = response.json()
            return Status(result["code"], result["message"]), None

        return Status(Status.UNEXPECTED_ERROR, "Response is empty"), None

    @handle_error()
    def preload_collection(self, table_name, partition_tags, timeout):
        url = self._uri + "/system/task"
        params = {"load": {"collection_name": table_name}}
        if partition_tags:
            params["load"]["partition_tags"] = partition_tags
        data = ujson.dumps(params)

        response = rq.put(url, data=data, timeout=timeout)

        if response.status_code == 200:
            return Status(message="Load successfuly")

        js = response.json()
        return Status(code=js["code"], message=js["message"])

    @handle_error()
    def reload_segments(self, collection_name, segment_ids, timeout=10):
        raise NotImplementedError("Not implemented in http server")

    @handle_error()
    def drop_collection(self, table_name, timeout):
        url = self._uri + "/collections/" + table_name
        response = rq.delete(url, timeout=timeout)
        if response.status_code == 204:
            return Status(message="Delete successfully!")

        js = response.json()
        return Status(js["code"], js["message"])

    @handle_error(returns=([],))
    def insert(self, table_name, records, ids, partition_tag, params, timeout, **kwargs):
        url = self._uri + "/collections/{}/vectors".format(table_name)

        data_dict = dict()
        if ids:
            data_dict["ids"] = list(map(str, ids))
        if partition_tag:
            data_dict["partition_tag"] = partition_tag

        if isinstance(records[0], bytes):
            vectors = [struct.unpack(str(len(r)) + 'B', r) for r in records]
            data_dict["vectors"] = vectors
        else:
            data_dict["vectors"] = records

        data = ujson.dumps(data_dict)

        headers = {"Content-Type": "application/json"}

        response = rq.post(url, data=data, headers=headers)
        js = response.json()

        if response.status_code == 201:
            ids = [int(item) for item in list(js["ids"])]
            return Status(message='Add vectors successfully!'), ids

        return Status(js["code"], js["message"]), []

    @handle_error(returns=(None,))
    def get_vectors_by_ids(self, collection_name, ids, timeout):
        status, table_schema = self.describe_collection(collection_name, timeout)
        if not status.OK():
            return status, None
        metric = table_schema.metric_type

        bin_vector = metric in list(MetricType.__members__.values())[3:]

        url = self._uri + "/collections/{}/vectors".format(collection_name)
        ids_list = list(map(str, ids))
        query_ids = ",".join(ids_list)
        url = url + "?ids=" + query_ids
        response = rq.get(url, timeout=timeout)
        result = response.json()

        if response.status_code == 200:
            vectors = result["vectors"]
            if not list(vectors):
                return Status(), []

            vector_results = []
            for vector_res in vectors:
                vector = list(vector_res["vector"])
                if bin_vector:
                    vector_results.append(bytes(vector))
                else:
                    vector_results.append(vector)
                    # return Status(),
            return Status(), vector_results

        return Status(result["code"], result["message"]), None

    @handle_error(returns=(None,))
    def get_vector_ids(self, table_name, segment_name, timeout):
        # TODO: here specify page_size hard is stupid, need server support query string 'qll_required'
        url = self._uri + "/collections/{}/segments/{}/ids?page_size=1000000".format(table_name, segment_name)
        response = rq.get(url, timeout=timeout)
        result = response.json()

        if response.status_code == 200:
            return Status(), list(map(int, result["ids"]))

        return Status(result["code"], result["message"]), None

    @handle_error()
    def create_index(self, table_name, index_type, index_params, timeout):
        url = self._uri + "/collections/{}/indexes".format(table_name)

        index = IndexValue2NameMap.get(index_type)
        request = dict()
        request["index_type"] = index
        request["params"] = index_params
        data = ujson.dumps(request)
        headers = {"Content-Type": "application/json"}

        response = rq.post(url, data=data, headers=headers, timeout=timeout)
        js = response.json()

        return Status(js["code"], js["message"])

    @handle_error(returns=(None,))
    def describe_index(self, table_name, timeout):
        url = self._uri + "/collections/{}/indexes".format(table_name)

        response = rq.get(url, timeout=timeout)

        if response.status_code >= 500:
            return Status(Status.UNEXPECTED_ERROR,
                          "Unexpected error.\n\tStatus code : {}, reason : {}"
                          .format(response.status_code, response.reason))

        js = response.json()

        if response.status_code == 200:
            index_type = IndexName2ValueMap.get(js["index_type"])
            return Status(), IndexParam(table_name, index_type, js["params"])

        return Status(js["code"], js["message"]), None

    @handle_error()
    def drop_index(self, table_name, timeout):
        url = self._uri + "/collections/{}/indexes".format(table_name)

        response = rq.delete(url)

        if response.status_code == 204:
            return Status()

        js = response.json()
        return Status(js["code"], js["message"])

    @handle_error()
    def create_partition(self, table_name, partition_tag, timeout=10):
        url = self._uri + "/collections/{}/partitions".format(table_name)

        data = ujson.dumps({"partition_tag": partition_tag})
        headers = {"Content-Type": "application/json"}

        response = rq.post(url, data=data, headers=headers, timeout=timeout)
        if response.status_code == 201:
            return Status()

        js = response.json()
        return Status(js["code"], js["message"])

    @handle_error(returns=([],))
    def show_partitions(self, table_name, timeout):
        url = self._uri + "/collections/{}/partitions".format(table_name)
        query_data = {"offset": 0, "page_size": 100}

        response = rq.get(url, params=query_data, timeout=timeout)
        if response.status_code >= 500:
            return Status(Status.UNEXPECTED_ERROR,
                          "Unexpected error. Status code : 500, reason: {}"
                          .format(response.reason)), None

        js = response.json()
        if response.status_code == 200:
            partition_list = \
                [PartitionParam(table_name, item["partition_tag"])
                 for item in js["partitions"]]

            return Status(), partition_list

        return Status(js["code"], js["message"]), []

    @handle_error(returns=(False,))
    def has_partition(self, collection_name, tag, timeout=30):
        url = self._uri + "/collections/{}/partitions".format(collection_name)

        response = rq.get(url, timeout=timeout)
        if response.status_code == 200:
            result = response.json()
            if result["count"] > 0:
                partitions = [p["partition_tag"] for p in list(result["partitions"])]
                return Status(), tag in partitions
            return Status(), False

        js = response.json()
        return Status(js["code"], js["message"]), False

    @handle_error()
    def drop_partition(self, collection_name, partition_tag, timeout=10):
        url = self._uri + "/collections/{}/partitions".format(collection_name)
        request = {
            "partition_tag": partition_tag
        }
        payload = ujson.dumps(request)

        response = rq.delete(url, data=payload, timeout=timeout)
        if response.status_code == 204:
            return Status()

        js = response.json()
        return Status(js["code"], js["message"])

    @handle_error(returns=(None,))
    def search(self, collection_name, top_k, query_records, partition_tags=None, search_params=None, timeout=None, **kwargs):
        url = self._uri + "/collections/{}/vectors".format(collection_name)

        search_body = dict()
        if partition_tags:
            search_body["partition_tags"] = partition_tags
        search_body["topk"] = top_k
        search_body["params"] = search_params

        if isinstance(query_records[0], bytes):
            vectors = [struct.unpack(str(len(r)) + 'B', r) for r in query_records]
            search_body["vectors"] = vectors
        else:
            vectors = query_records
            search_body["vectors"] = vectors

        data = ujson.dumps({"search": search_body})
        headers = {"Content-Type": "application/json"}

        response = rq.put(url, data, headers=headers)

        if response.status_code == 200:
            return Status(), TopKQueryResult2(response)

        js = response.json()
        return Status(js["code"], js["message"]), None

    @handle_error(returns=(None,))
    def search_by_ids(self, collection_name, ids, top_k, partition_tags=None, search_params=None, timeout=None, **kwargs):
        url = self._uri + "/collections/{}/vectors".format(collection_name)
        body_dict = dict()
        body_dict["topk"] = top_k
        body_dict["ids"] = list(map(str, ids))
        if partition_tags:
            body_dict["partition_tags"] = partition_tags
        if search_params:
            body_dict["params"] = search_params

        data = ujson.dumps({"search": body_dict})
        headers = {"Content-Type": "application/json"}

        response = rq.put(url, data, headers=headers, timeout=timeout)

        if response.status_code == 200:
            return Status(), TopKQueryResult2(response)

        js = response.json()
        return Status(js["code"], js["message"]), None

    @handle_error(returns=(None,))
    def search_in_files(self, collection_name, file_ids, query_records, top_k, search_params, timeout, **kwargs):
        url = self._uri + "/collections/{}/vectors".format(collection_name)

        body_dict = dict()
        body_dict["topk"] = top_k
        body_dict["file_ids"] = list(map(str, file_ids))
        body_dict["params"] = search_params

        if isinstance(query_records[0], bytes):
            vectors = [struct.unpack(str(len(r)) + 'B', r) for r in query_records]
            body_dict["vectors"] = vectors
        else:
            vectors = query_records
            body_dict["vectors"] = vectors

        data = ujson.dumps({"search": body_dict})
        headers = {"Content-Type": "application/json"}

        response = rq.put(url, data, headers=headers, timeout=timeout)

        if response.status_code == 200:
            return Status(), TopKQueryResult2(response)

        js = response.json()
        return Status(js["code"], js["message"]), None

    @handle_error()
    def delete_by_id(self, table_name, id_array, timeout=None):
        url = self._uri + "/collections/{}/vectors".format(table_name)
        headers = {"Content-Type": "application/json"}
        ids = list(map(str, id_array))
        request = {"delete": {"ids": ids}}

        response = rq.put(url, data=ujson.dumps(request), headers=headers, timeout=timeout)
        result = response.json()
        return Status(result["code"], result["message"])

    @handle_error()
    def flush(self, table_name_array, *kwargs):
        url = self._uri + "/system/task"
        headers = {"Content-Type": "application/json"}
        request = {"flush": {"collection_names": table_name_array}}

        response = rq.put(url, ujson.dumps(request), headers=headers)
        result = response.json()
        return Status(result["code"], result["message"])

    @handle_error()
    def compact(self, table_name, timeout):
        url = self._uri + "/system/task"
        headers = {"Content-Type": "application/json"}
        request = {"compact": {"collection_name": table_name}}

        response = rq.put(url, ujson.dumps(request), headers=headers)
        result = response.json()
        return Status(result["code"], result["message"])
